"""
This type stub file was generated by pyright.
"""

from contextlib import ContextDecorator, contextmanager
from typing import Any, Callable, Generic, Iterable, Iterator, NamedTuple
from pandas._typing import T

"""
The config module holds package-wide configurables and provides
a uniform API for working with them.

Overview
========

This module supports the following requirements:
- options are referenced using keys in dot.notation, e.g. "x.y.option - z".
- keys are case-insensitive.
- functions should accept partial/regex keys, when unambiguous.
- options can be registered by modules at import time.
- options can be registered at init-time (via core.config_init)
- options have a default value, and (optionally) a description and
  validation function associated with them.
- options can be deprecated, in which case referencing them
  should produce a warning.
- deprecated options can optionally be rerouted to a replacement
  so that accessing a deprecated option reroutes to a differently
  named option.
- options can be reset to their default value.
- all option can be reset to their default value at once.
- all options in a certain sub - namespace can be reset at once.
- the user can set / get / reset or ask for the description of an option.
- a developer can register and mark an option as deprecated.
- you can register a callback to be invoked when the option value
  is set or reset. Changing the stored value is considered misuse, but
  is not verboten.

Implementation
==============

- Data is stored using nested dictionaries, and should be accessed
  through the provided API.

- "Registered options" and "Deprecated options" have metadata associated
  with them, which are stored in auxiliary dictionaries keyed on the
  fully-qualified key, e.g. "x.y.z.option".

- the config_init module is imported by the package's __init__.py file.
  placing any register_option() calls there will ensure those options
  are available as soon as pandas is loaded. If you use register_option
  in a module, it will only be available after that module is imported,
  which you should be aware of.

- `config_prefix` is a context_manager (for use with the `with` keyword)
  which can save developers some typing, see the docstring.

"""
class DeprecatedOption(NamedTuple):
    key: str
    msg: str | None
    rkey: str | None
    removal_ver: str | None
    ...


class RegisteredOption(NamedTuple):
    key: str
    defval: object
    doc: str
    validator: Callable[[object], Any] | None
    cb: Callable[[str], Any] | None
    ...


_deprecated_options: dict[str, DeprecatedOption] = ...
_registered_options: dict[str, RegisteredOption] = ...
_global_config: dict[str, Any] = ...
_reserved_keys: list[str] = ...
class OptionError(AttributeError, KeyError):
    """
    Exception raised for pandas.options.

    Backwards compatible with KeyError checks.
    """
    ...


def get_default_val(pat: str): # -> object:
    ...

class DictWrapper:
    """provide attribute-style access to a nested dict"""
    def __init__(self, d: dict[str, Any], prefix: str = ...) -> None:
        ...
    
    def __setattr__(self, key: str, val: Any) -> None:
        ...
    
    def __getattr__(self, key: str): # -> DictWrapper | Any:
        ...
    
    def __dir__(self) -> Iterable[str]:
        ...
    


class CallableDynamicDoc(Generic[T]):
    def __init__(self, func: Callable[..., T], doc_tmpl: str) -> None:
        ...
    
    def __call__(self, *args, **kwds) -> T:
        ...
    
    @property
    def __doc__(self) -> str:
        ...
    


_get_option_tmpl = ...
_set_option_tmpl = ...
_describe_option_tmpl = ...
_reset_option_tmpl = ...
get_option = ...
set_option = ...
reset_option = ...
describe_option = ...
options = ...
class option_context(ContextDecorator):
    """
    Context manager to temporarily set options in the `with` statement context.

    You need to invoke as ``option_context(pat, val, [(pat, val), ...])``.

    Examples
    --------
    >>> with option_context('display.max_rows', 10, 'display.max_columns', 5):
    ...     pass
    """
    def __init__(self, *args) -> None:
        ...
    
    def __enter__(self) -> None:
        ...
    
    def __exit__(self, *args) -> None:
        ...
    


def register_option(key: str, defval: object, doc: str = ..., validator: Callable[[object], Any] | None = ..., cb: Callable[[str], Any] | None = ...) -> None:
    """
    Register an option in the package-wide pandas config object

    Parameters
    ----------
    key : str
        Fully-qualified key, e.g. "x.y.option - z".
    defval : object
        Default value of the option.
    doc : str
        Description of the option.
    validator : Callable, optional
        Function of a single argument, should raise `ValueError` if
        called with a value which is not a legal value for the option.
    cb
        a function of a single argument "key", which is called
        immediately after an option value is set/reset. key is
        the full name of the option.

    Raises
    ------
    ValueError if `validator` is specified and `defval` is not a valid value.

    """
    ...

def deprecate_option(key: str, msg: str | None = ..., rkey: str | None = ..., removal_ver: str | None = ...) -> None:
    """
    Mark option `key` as deprecated, if code attempts to access this option,
    a warning will be produced, using `msg` if given, or a default message
    if not.
    if `rkey` is given, any access to the key will be re-routed to `rkey`.

    Neither the existence of `key` nor that if `rkey` is checked. If they
    do not exist, any subsequence access will fail as usual, after the
    deprecation warning is given.

    Parameters
    ----------
    key : str
        Name of the option to be deprecated.
        must be a fully-qualified option name (e.g "x.y.z.rkey").
    msg : str, optional
        Warning message to output when the key is referenced.
        if no message is given a default message will be emitted.
    rkey : str, optional
        Name of an option to reroute access to.
        If specified, any referenced `key` will be
        re-routed to `rkey` including set/get/reset.
        rkey must be a fully-qualified option name (e.g "x.y.z.rkey").
        used by the default message if no `msg` is specified.
    removal_ver : str, optional
        Specifies the version in which this option will
        be removed. used by the default message if no `msg` is specified.

    Raises
    ------
    OptionError
        If the specified key has already been deprecated.
    """
    ...

def pp_options_list(keys: Iterable[str], width=..., _print: bool = ...): # -> str | None:
    """Builds a concise listing of available options, grouped by prefix"""
    ...

@contextmanager
def config_prefix(prefix) -> Iterator[None]:
    """
    contextmanager for multiple invocations of API with a common prefix

    supported API functions: (register / get / set )__option

    Warning: This is not thread - safe, and won't work properly if you import
    the API functions into your module using the "from x import y" construct.

    Example
    -------
    import pandas._config.config as cf
    with cf.config_prefix("display.font"):
        cf.register_option("color", "red")
        cf.register_option("size", " 5 pt")
        cf.set_option(size, " 6 pt")
        cf.get_option(size)
        ...

        etc'

    will register options "display.font.color", "display.font.size", set the
    value of "display.font.size"... and so on.
    """
    ...

def is_type_factory(_type: type[Any]) -> Callable[[Any], None]:
    """

    Parameters
    ----------
    `_type` - a type to be compared against (e.g. type(x) == `_type`)

    Returns
    -------
    validator - a function of a single argument x , which raises
                ValueError if type(x) is not equal to `_type`

    """
    ...

def is_instance_factory(_type) -> Callable[[Any], None]:
    """

    Parameters
    ----------
    `_type` - the type to be checked against

    Returns
    -------
    validator - a function of a single argument x , which raises
                ValueError if x is not an instance of `_type`

    """
    ...

def is_one_of_factory(legal_values) -> Callable[[Any], None]:
    ...

def is_nonnegative_int(value: object) -> None:
    """
    Verify that value is None or a positive int.

    Parameters
    ----------
    value : None or int
            The `value` to be checked.

    Raises
    ------
    ValueError
        When the value is not None or is a negative integer
    """
    ...

is_int = ...
is_bool = ...
is_float = ...
is_str = ...
is_text = ...
def is_callable(obj) -> bool:
    """

    Parameters
    ----------
    `obj` - the object to be checked

    Returns
    -------
    validator - returns True if object is callable
        raises ValueError otherwise.

    """
    ...




"""
This type stub file was generated by pyright.
"""

import numpy as np
import pyarrow as pa
from pandas.compat import pa_version_under1p01
from pandas.util._decorators import cache_readonly
from pandas.core.dtypes.base import StorageExtensionDtype, register_extension_dtype

if notpa_version_under1p01:
    ...
@register_extension_dtype
class ArrowDtype(StorageExtensionDtype):
    """
    An ExtensionDtype for PyArrow data types.

    .. warning::

       ArrowDtype is considered experimental. The implementation and
       parts of the API may change without warning.

    While most ``dtype`` arguments can accept the "string"
    constructor, e.g. ``"int64[pyarrow]"``, ArrowDtype is useful
    if the data type contains parameters like ``pyarrow.timestamp``.

    Parameters
    ----------
    pyarrow_dtype : pa.DataType
        An instance of a `pyarrow.DataType <https://arrow.apache.org/docs/python/api/datatypes.html#factory-functions>`__.

    Attributes
    ----------
    pyarrow_dtype

    Methods
    -------
    None

    Returns
    -------
    ArrowDtype

    Examples
    --------
    >>> import pyarrow as pa
    >>> pd.ArrowDtype(pa.int64())
    int64[pyarrow]

    Types with parameters must be constructed with ArrowDtype.

    >>> pd.ArrowDtype(pa.timestamp("s", tz="America/New_York"))
    timestamp[s, tz=America/New_York][pyarrow]
    >>> pd.ArrowDtype(pa.list_(pa.int64()))
    list<item: int64>[pyarrow]
    """
    _metadata = ...
    def __init__(self, pyarrow_dtype: pa.DataType) -> None:
        ...
    
    def __repr__(self) -> str:
        ...
    
    @property
    def type(self): # -> Any:
        """
        Returns pyarrow.DataType.
        """
        ...
    
    @property
    def name(self) -> str:
        """
        A string identifying the data type.
        """
        ...
    
    @cache_readonly
    def numpy_dtype(self) -> np.dtype:
        """Return an instance of the related numpy dtype"""
        ...
    
    @cache_readonly
    def kind(self) -> str:
        ...
    
    @cache_readonly
    def itemsize(self) -> int:
        """Return the number of bytes in this dtype"""
        ...
    
    @classmethod
    def construct_array_type(cls): # -> Type[ArrowExtensionArray]:
        """
        Return the array type associated with this dtype.

        Returns
        -------
        type
        """
        ...
    
    @classmethod
    def construct_from_string(cls, string: str) -> ArrowDtype:
        """
        Construct this type from a string.

        Parameters
        ----------
        string : str
            string should follow the format f"{pyarrow_type}[pyarrow]"
            e.g. int64[pyarrow]
        """
        ...
    
    def __from_arrow__(self, array: pa.Array | pa.ChunkedArray): # -> ArrowExtensionArray:
        """
        Construct IntegerArray/FloatingArray from pyarrow Array/ChunkedArray.
        """
        ...
    





"""
This type stub file was generated by pyright.
"""

import numpy as np
from typing import Hashable
from pandas._libs.tslibs import BaseOffset
from pandas._typing import Dtype, npt
from pandas.util._decorators import doc
from pandas.core.dtypes.dtypes import PeriodDtype
from pandas.core.arrays.period import PeriodArray
from pandas.core.indexes.datetimelike import DatetimeIndexOpsMixin
from pandas.core.indexes.datetimes import DatetimeIndex, Index
from pandas.core.indexes.extension import inherit_names
from pandas.core.indexes.numeric import Int64Index

_index_doc_kwargs = ...
_shared_doc_kwargs = ...
@inherit_names(["strftime", "start_time", "end_time"] + PeriodArray._field_ops, PeriodArray, wrap=True)
@inherit_names(["is_leap_year", "_format_native_types"], PeriodArray)
class PeriodIndex(DatetimeIndexOpsMixin):
    """
    Immutable ndarray holding ordinal values indicating regular periods in time.

    Index keys are boxed to Period objects which carries the metadata (eg,
    frequency information).

    Parameters
    ----------
    data : array-like (1d int np.ndarray or PeriodArray), optional
        Optional period-like data to construct index with.
    copy : bool
        Make a copy of input ndarray.
    freq : str or period object, optional
        One of pandas period strings or corresponding objects.
    year : int, array, or Series, default None
    month : int, array, or Series, default None
    quarter : int, array, or Series, default None
    day : int, array, or Series, default None
    hour : int, array, or Series, default None
    minute : int, array, or Series, default None
    second : int, array, or Series, default None
    dtype : str or PeriodDtype, default None

    Attributes
    ----------
    day
    dayofweek
    day_of_week
    dayofyear
    day_of_year
    days_in_month
    daysinmonth
    end_time
    freq
    freqstr
    hour
    is_leap_year
    minute
    month
    quarter
    qyear
    second
    start_time
    week
    weekday
    weekofyear
    year

    Methods
    -------
    asfreq
    strftime
    to_timestamp

    See Also
    --------
    Index : The base pandas Index type.
    Period : Represents a period of time.
    DatetimeIndex : Index with datetime64 data.
    TimedeltaIndex : Index of timedelta64 data.
    period_range : Create a fixed-frequency PeriodIndex.

    Examples
    --------
    >>> idx = pd.PeriodIndex(year=[2000, 2002], quarter=[1, 3])
    >>> idx
    PeriodIndex(['2000Q1', '2002Q3'], dtype='period[Q-DEC]')
    """
    _typ = ...
    _data: PeriodArray
    freq: BaseOffset
    dtype: PeriodDtype
    _data_cls = PeriodArray
    _supports_partial_string_indexing = ...
    @doc(PeriodArray.asfreq, other="pandas.arrays.PeriodArray", other_name="PeriodArray", **_shared_doc_kwargs)
    def asfreq(self, freq=..., how: str = ...) -> PeriodIndex:
        ...
    
    @doc(PeriodArray.to_timestamp)
    def to_timestamp(self, freq=..., how: str = ...) -> DatetimeIndex:
        ...
    
    @property
    @doc(PeriodArray.hour.fget)
    def hour(self) -> Int64Index:
        ...
    
    @property
    @doc(PeriodArray.minute.fget)
    def minute(self) -> Int64Index:
        ...
    
    @property
    @doc(PeriodArray.second.fget)
    def second(self) -> Int64Index:
        ...
    
    def __new__(cls, data=..., ordinal=..., freq=..., dtype: Dtype | None = ..., copy: bool = ..., name: Hashable = ..., **fields) -> PeriodIndex:
        ...
    
    @property
    def values(self) -> np.ndarray:
        ...
    
    def asof_locs(self, where: Index, mask: npt.NDArray[np.bool_]) -> np.ndarray:
        """
        where : array of timestamps
        mask : np.ndarray[bool]
            Array of booleans where data is not NA.
        """
        ...
    
    @doc(Index.astype)
    def astype(self, dtype, copy: bool = ..., how=...): # -> DatetimeIndex | PeriodIndex | Index:
        ...
    
    @property
    def is_full(self) -> bool:
        """
        Returns True if this PeriodIndex is range-like in that all Periods
        between start and end are present, in order.
        """
        ...
    
    @property
    def inferred_type(self) -> str:
        ...
    
    def get_loc(self, key, method=..., tolerance=...): # -> slice | ndarray[Any, _dtype] | int | ndarray[Unknown, Unknown]:
        """
        Get integer location for requested label.

        Parameters
        ----------
        key : Period, NaT, str, or datetime
            String or datetime key must be parsable as Period.

        Returns
        -------
        loc : int or ndarray[int64]

        Raises
        ------
        KeyError
            Key is not present in the index.
        TypeError
            If key is listlike or otherwise not hashable.
        """
        ...
    


def period_range(start=..., end=..., periods: int | None = ..., freq=..., name=...) -> PeriodIndex:
    """
    Return a fixed frequency PeriodIndex.

    The day (calendar) is the default frequency.

    Parameters
    ----------
    start : str or period-like, default None
        Left bound for generating periods.
    end : str or period-like, default None
        Right bound for generating periods.
    periods : int, default None
        Number of periods to generate.
    freq : str or DateOffset, optional
        Frequency alias. By default the freq is taken from `start` or `end`
        if those are Period objects. Otherwise, the default is ``"D"`` for
        daily frequency.
    name : str, default None
        Name of the resulting PeriodIndex.

    Returns
    -------
    PeriodIndex

    Notes
    -----
    Of the three parameters: ``start``, ``end``, and ``periods``, exactly two
    must be specified.

    To learn more about the frequency strings, please see `this link
    <https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#offset-aliases>`__.

    Examples
    --------
    >>> pd.period_range(start='2017-01-01', end='2018-01-01', freq='M')
    PeriodIndex(['2017-01', '2017-02', '2017-03', '2017-04', '2017-05', '2017-06',
             '2017-07', '2017-08', '2017-09', '2017-10', '2017-11', '2017-12',
             '2018-01'],
            dtype='period[M]')

    If ``start`` or ``end`` are ``Period`` objects, they will be used as anchor
    endpoints for a ``PeriodIndex`` with frequency matching that of the
    ``period_range`` constructor.

    >>> pd.period_range(start=pd.Period('2017Q1', freq='Q'),
    ...                 end=pd.Period('2017Q2', freq='Q'), freq='M')
    PeriodIndex(['2017-03', '2017-04', '2017-05', '2017-06'],
                dtype='period[M]')
    """
    ...




"""
This type stub file was generated by pyright.
"""

import numpy as np
from typing import Any, Callable, TYPE_CHECKING, TypeVar
from pandas._libs.tslibs import BaseOffset, Resolution
from pandas.util._decorators import Appender, cache_readonly, doc
from pandas.core.arrays import DatetimeArray, PeriodArray, TimedeltaArray
from pandas.core.arrays.datetimelike import DatetimeLikeArrayMixin
from pandas.core.indexes.base import Index, _index_shared_docs
from pandas.core.indexes.extension import NDArrayBackedExtensionIndex, inherit_names

"""
Base and utility classes for tseries type pandas objects.
"""
if TYPE_CHECKING:
    ...
_index_doc_kwargs = ...
_T = TypeVar("_T", bound="DatetimeIndexOpsMixin")
_TDT = TypeVar("_TDT", bound="DatetimeTimedeltaMixin")
@inherit_names(["inferred_freq", "_resolution_obj", "resolution"], DatetimeLikeArrayMixin, cache=True)
@inherit_names(["mean", "asi8", "freq", "freqstr"], DatetimeLikeArrayMixin)
class DatetimeIndexOpsMixin(NDArrayBackedExtensionIndex):
    """
    Common ops mixin to support a unified interface datetimelike Index.
    """
    _is_numeric_dtype = ...
    _can_hold_strings = ...
    _data: DatetimeArray | TimedeltaArray | PeriodArray
    freq: BaseOffset | None
    freqstr: str | None
    _resolution_obj: Resolution
    @cache_readonly
    def hasnans(self) -> bool:
        ...
    
    def equals(self, other: Any) -> bool:
        """
        Determines if two Index objects contain the same elements.
        """
        ...
    
    @Appender(Index.__contains__.__doc__)
    def __contains__(self, key: Any) -> bool:
        ...
    
    def format(self, name: bool = ..., formatter: Callable | None = ..., na_rep: str = ..., date_format: str | None = ...) -> list[str]:
        """
        Render a string representation of the Index.
        """
        ...
    
    def shift(self: _T, periods: int = ..., freq=...) -> _T:
        """
        Shift index by desired number of time frequency increments.

        This method is for shifting the values of datetime-like indexes
        by a specified time increment a given number of times.

        Parameters
        ----------
        periods : int, default 1
            Number of periods (or increments) to shift by,
            can be positive or negative.
        freq : pandas.DateOffset, pandas.Timedelta or string, optional
            Frequency increment to shift by.
            If None, the index is shifted by its own `freq` attribute.
            Offset aliases are valid strings, e.g., 'D', 'W', 'M' etc.

        Returns
        -------
        pandas.DatetimeIndex
            Shifted index.

        See Also
        --------
        Index.shift : Shift values of Index.
        PeriodIndex.shift : Shift values of PeriodIndex.
        """
        ...
    


class DatetimeTimedeltaMixin(DatetimeIndexOpsMixin):
    """
    Mixin class for methods shared by DatetimeIndex and TimedeltaIndex,
    but not PeriodIndex
    """
    _data: DatetimeArray | TimedeltaArray
    _comparables = ...
    _attributes = ...
    _is_monotonic_increasing = ...
    _is_monotonic_decreasing = ...
    _is_unique = ...
    _join_precedence = ...
    def is_type_compatible(self, kind: str) -> bool:
        ...
    
    @property
    def values(self) -> np.ndarray:
        ...
    
    @doc(NDArrayBackedExtensionIndex.delete)
    def delete(self, loc) -> DatetimeTimedeltaMixin:
        ...
    
    @doc(NDArrayBackedExtensionIndex.insert)
    def insert(self, loc: int, item): # -> Self@DatetimeTimedeltaMixin | Index:
        ...
    
    @Appender(_index_shared_docs["take"] % _index_doc_kwargs)
    def take(self, indices, axis=..., allow_fill=..., fill_value=..., **kwargs): # -> NDArrayBackedExtensionIndex:
        ...
    





"""
This type stub file was generated by pyright.
"""

from typing import Literal, TypeVar, final
from pandas._typing import ArrayLike, DtypeObj, Shape
from pandas.core.base import PandasObject
from pandas.core.indexes.api import Index

"""
Base class for the internal managers. Both BlockManager and ArrayManager
inherit from this class.
"""
T = TypeVar("T", bound="DataManager")
class DataManager(PandasObject):
    axes: list[Index]
    @property
    def items(self) -> Index:
        ...
    
    @final
    def __len__(self) -> int:
        ...
    
    @property
    def ndim(self) -> int:
        ...
    
    @property
    def shape(self) -> Shape:
        ...
    
    def reindex_indexer(self: T, new_axis, indexer, axis: int, fill_value=..., allow_dups: bool = ..., copy: bool = ..., only_slice: bool = ...) -> T:
        ...
    
    @final
    def reindex_axis(self: T, new_index: Index, axis: int, fill_value=..., only_slice: bool = ...) -> T:
        """
        Conform data manager to new index.
        """
        ...
    
    @final
    def equals(self, other: object) -> bool:
        """
        Implementation for DataFrame.equals
        """
        ...
    
    def apply(self: T, f, align_keys: list[str] | None = ..., ignore_failures: bool = ..., **kwargs) -> T:
        ...
    
    @final
    def isna(self: T, func) -> T:
        ...
    
    def is_consolidated(self) -> bool:
        ...
    
    def consolidate(self: T) -> T:
        ...
    


class SingleDataManager(DataManager):
    @property
    def ndim(self) -> Literal[1]:
        ...
    
    @final
    @property
    def array(self) -> ArrayLike:
        """
        Quick access to the backing array of the Block or SingleArrayManager.
        """
        ...
    
    def setitem_inplace(self, indexer, value) -> None:
        """
        Set values with indexer.

        For Single[Block/Array]Manager, this backs s[indexer] = value

        This is an inplace version of `setitem()`, mutating the manager/values
        in place, not returning a new Manager (and Block), and thus never changing
        the dtype.
        """
        ...
    
    def grouped_reduce(self, func, ignore_failures: bool = ...):
        """
        ignore_failures : bool, default False
            Not used; for compatibility with ArrayManager/BlockManager.
        """
        ...
    
    @classmethod
    def from_array(cls, arr: ArrayLike, index: Index):
        ...
    


def interleaved_dtype(dtypes: list[DtypeObj]) -> DtypeObj | None:
    """
    Find the common dtype for `blocks`.

    Parameters
    ----------
    blocks : List[DtypeObj]

    Returns
    -------
    dtype : np.dtype, ExtensionDtype, or None
        None is returned when `blocks` is empty.
    """
    ...




"""
This type stub file was generated by pyright.
"""

from typing import Any, TYPE_CHECKING
from pandas._typing import FilePath, StorageOptions, WriteExcelBuffer
from pandas.io.excel._base import ExcelWriter

if TYPE_CHECKING:
    ...
class ODSWriter(ExcelWriter):
    _engine = ...
    _supported_extensions = ...
    def __init__(self, path: FilePath | WriteExcelBuffer | ExcelWriter, engine: str | None = ..., date_format: str | None = ..., datetime_format=..., mode: str = ..., storage_options: StorageOptions = ..., if_sheet_exists: str | None = ..., engine_kwargs: dict[str, Any] | None = ..., **kwargs) -> None:
        ...
    
    @property
    def book(self):
        """
        Book instance of class odf.opendocument.OpenDocumentSpreadsheet.

        This attribute can be used to access engine-specific features.
        """
        ...
    
    @property
    def sheets(self) -> dict[str, Any]:
        """Mapping of sheet names to sheet objects."""
        ...
    





"""
This type stub file was generated by pyright.
"""

from typing import Generic, TypeVar
from pandas.core.indexing import IndexingMixin

_IndexingMixinT = TypeVar("_IndexingMixinT", bound=IndexingMixin)
class NDFrameIndexerBase(Generic[_IndexingMixinT]):
    name: str
    obj: _IndexingMixinT
    def __init__(self, name: str, obj: _IndexingMixinT) -> None:
        ...
    
    @property
    def ndim(self) -> int:
        ...
    





"""
This type stub file was generated by pyright.
"""

def generate_online_numba_ewma_func(nopython: bool, nogil: bool, parallel: bool): # -> (values: ndarray[Unknown, Unknown], deltas: ndarray[Unknown, Unknown], minimum_periods: int, old_wt_factor: float, new_wt: float, old_wt: ndarray[Unknown, Unknown], adjust: bool, ignore_na: bool) -> tuple[NDArray[float64], ndarray[Unknown, Unknown]]:
    """
    Generate a numba jitted groupby ewma function specified by values
    from engine_kwargs.

    Parameters
    ----------
    nopython : bool
        nopython to be passed into numba.jit
    nogil : bool
        nogil to be passed into numba.jit
    parallel : bool
        parallel to be passed into numba.jit

    Returns
    -------
    Numba function
    """
    ...

class EWMMeanState:
    def __init__(self, com, adjust, ignore_na, axis, shape) -> None:
        ...
    
    def run_ewm(self, weighted_avg, deltas, min_periods, ewm_func):
        ...
    
    def reset(self) -> None:
        ...
    





"""
This type stub file was generated by pyright.
"""

import numpy as np
from pandas.core.dtypes.dtypes import register_extension_dtype
from pandas.core.arrays.numeric import NumericArray, NumericDtype

class FloatingDtype(NumericDtype):
    """
    An ExtensionDtype to hold a single size of floating dtype.

    These specific implementations are subclasses of the non-public
    FloatingDtype. For example we have Float32Dtype to represent float32.

    The attributes name & type are set when these subclasses are created.
    """
    _default_np_dtype = ...
    _checker = ...
    @classmethod
    def construct_array_type(cls) -> type[FloatingArray]:
        """
        Return the array type associated with this dtype.

        Returns
        -------
        type
        """
        ...
    


class FloatingArray(NumericArray):
    """
    Array of floating (optional missing) values.

    .. versionadded:: 1.2.0

    .. warning::

       FloatingArray is currently experimental, and its API or internal
       implementation may change without warning. Especially the behaviour
       regarding NaN (distinct from NA missing values) is subject to change.

    We represent a FloatingArray with 2 numpy arrays:

    - data: contains a numpy float array of the appropriate dtype
    - mask: a boolean array holding a mask on the data, True is missing

    To construct an FloatingArray from generic array-like input, use
    :func:`pandas.array` with one of the float dtypes (see examples).

    See :ref:`integer_na` for more.

    Parameters
    ----------
    values : numpy.ndarray
        A 1-d float-dtype array.
    mask : numpy.ndarray
        A 1-d boolean-dtype array indicating missing values.
    copy : bool, default False
        Whether to copy the `values` and `mask`.

    Attributes
    ----------
    None

    Methods
    -------
    None

    Returns
    -------
    FloatingArray

    Examples
    --------
    Create an FloatingArray with :func:`pandas.array`:

    >>> pd.array([0.1, None, 0.3], dtype=pd.Float32Dtype())
    <FloatingArray>
    [0.1, <NA>, 0.3]
    Length: 3, dtype: Float32

    String aliases for the dtypes are also available. They are capitalized.

    >>> pd.array([0.1, None, 0.3], dtype="Float32")
    <FloatingArray>
    [0.1, <NA>, 0.3]
    Length: 3, dtype: Float32
    """
    _dtype_cls = FloatingDtype
    _internal_fill_value = ...
    _truthy_value = ...
    _falsey_value = ...


_dtype_docstring = ...
@register_extension_dtype
class Float32Dtype(FloatingDtype):
    type = np.float32
    name = ...
    __doc__ = ...


@register_extension_dtype
class Float64Dtype(FloatingDtype):
    type = np.float64
    name = ...
    __doc__ = ...


FLOAT_STR_TO_DTYPE = ...



"""
This type stub file was generated by pyright.
"""

from typing import Any, TYPE_CHECKING
from pandas._typing import Dtype, type_t
from pandas.core.dtypes.base import ExtensionDtype, register_extension_dtype
from pandas.core.arrays.sparse.array import SparseArray

"""Sparse Dtype"""
if TYPE_CHECKING:
    ...
@register_extension_dtype
class SparseDtype(ExtensionDtype):
    """
    Dtype for data stored in :class:`SparseArray`.

    This dtype implements the pandas ExtensionDtype interface.

    Parameters
    ----------
    dtype : str, ExtensionDtype, numpy.dtype, type, default numpy.float64
        The dtype of the underlying array storing the non-fill value values.
    fill_value : scalar, optional
        The scalar value not stored in the SparseArray. By default, this
        depends on `dtype`.

        =========== ==========
        dtype       na_value
        =========== ==========
        float       ``np.nan``
        int         ``0``
        bool        ``False``
        datetime64  ``pd.NaT``
        timedelta64 ``pd.NaT``
        =========== ==========

        The default value may be overridden by specifying a `fill_value`.

    Attributes
    ----------
    None

    Methods
    -------
    None
    """
    _metadata = ...
    def __init__(self, dtype: Dtype = ..., fill_value: Any = ...) -> None:
        ...
    
    def __hash__(self) -> int:
        ...
    
    def __eq__(self, other: Any) -> bool:
        ...
    
    @property
    def fill_value(self): # -> Any:
        """
        The fill value of the array.

        Converting the SparseArray to a dense ndarray will fill the
        array with this value.

        .. warning::

           It's possible to end up with a SparseArray that has ``fill_value``
           values in ``sp_values``. This can occur, for example, when setting
           ``SparseArray.fill_value`` directly.
        """
        ...
    
    @property
    def kind(self) -> str:
        """
        The sparse kind. Either 'integer', or 'block'.
        """
        ...
    
    @property
    def type(self): # -> Type[object_] | type:
        ...
    
    @property
    def subtype(self): # -> DtypeObj | dtype[object_]:
        ...
    
    @property
    def name(self) -> str:
        ...
    
    def __repr__(self) -> str:
        ...
    
    @classmethod
    def construct_array_type(cls) -> type_t[SparseArray]:
        """
        Return the array type associated with this dtype.

        Returns
        -------
        type
        """
        ...
    
    @classmethod
    def construct_from_string(cls, string: str) -> SparseDtype:
        """
        Construct a SparseDtype from a string form.

        Parameters
        ----------
        string : str
            Can take the following forms.

            string           dtype
            ================ ============================
            'int'            SparseDtype[np.int64, 0]
            'Sparse'         SparseDtype[np.float64, nan]
            'Sparse[int]'    SparseDtype[np.int64, 0]
            'Sparse[int, 0]' SparseDtype[np.int64, 0]
            ================ ============================

            It is not possible to specify non-default fill values
            with a string. An argument like ``'Sparse[int, 1]'``
            will raise a ``TypeError`` because the default fill value
            for integers is 0.

        Returns
        -------
        SparseDtype
        """
        ...
    
    @classmethod
    def is_dtype(cls, dtype: object) -> bool:
        ...
    
    def update_dtype(self, dtype) -> SparseDtype:
        """
        Convert the SparseDtype to a new dtype.

        This takes care of converting the ``fill_value``.

        Parameters
        ----------
        dtype : Union[str, numpy.dtype, SparseDtype]
            The new dtype to use.

            * For a SparseDtype, it is simply returned
            * For a NumPy dtype (or str), the current fill value
              is converted to the new dtype, and a SparseDtype
              with `dtype` and the new fill value is returned.

        Returns
        -------
        SparseDtype
            A new SparseDtype with the correct `dtype` and fill value
            for that `dtype`.

        Raises
        ------
        ValueError
            When the current fill value cannot be converted to the
            new `dtype` (e.g. trying to convert ``np.nan`` to an
            integer dtype).


        Examples
        --------
        >>> SparseDtype(int, 0).update_dtype(float)
        Sparse[float64, 0.0]

        >>> SparseDtype(int, 1).update_dtype(SparseDtype(float, np.nan))
        Sparse[float64, nan]
        """
        ...
    





"""
This type stub file was generated by pyright.
"""

import numpy as np
from pandas._typing import ArrayLike, npt

def quantile_compat(values: ArrayLike, qs: npt.NDArray[np.float64], interpolation: str) -> ArrayLike:
    """
    Compute the quantiles of the given values for each quantile in `qs`.

    Parameters
    ----------
    values : np.ndarray or ExtensionArray
    qs : np.ndarray[float64]
    interpolation : str

    Returns
    -------
    np.ndarray or ExtensionArray
    """
    ...

def quantile_with_mask(values: np.ndarray, mask: npt.NDArray[np.bool_], fill_value, qs: npt.NDArray[np.float64], interpolation: str) -> np.ndarray:
    """
    Compute the quantiles of the given values for each quantile in `qs`.

    Parameters
    ----------
    values : np.ndarray
        For ExtensionArray, this is _values_for_factorize()[0]
    mask : np.ndarray[bool]
        mask = isna(values)
        For ExtensionArray, this is computed before calling _value_for_factorize
    fill_value : Scalar
        The value to interpret fill NA entries with
        For ExtensionArray, this is _values_for_factorize()[1]
    qs : np.ndarray[float64]
    interpolation : str
        Type of interpolation

    Returns
    -------
    np.ndarray

    Notes
    -----
    Assumes values is already 2D.  For ExtensionArray this means np.atleast_2d
    has been called on _values_for_factorize()[0]

    Quantile is computed along axis=1.
    """
    ...




"""
This type stub file was generated by pyright.
"""

from typing import Literal
from pandas._libs.lib import NoDefault
from pandas import Index, Series

def assert_almost_equal(left, right, check_dtype: bool | Literal["equiv"] = ..., check_less_precise: bool | int | NoDefault = ..., rtol: float = ..., atol: float = ..., **kwargs) -> None:
    """
    Check that the left and right objects are approximately equal.

    By approximately equal, we refer to objects that are numbers or that
    contain numbers which may be equivalent to specific levels of precision.

    Parameters
    ----------
    left : object
    right : object
    check_dtype : bool or {'equiv'}, default 'equiv'
        Check dtype if both a and b are the same type. If 'equiv' is passed in,
        then `RangeIndex` and `Int64Index` are also considered equivalent
        when doing type checking.
    check_less_precise : bool or int, default False
        Specify comparison precision. 5 digits (False) or 3 digits (True)
        after decimal points are compared. If int, then specify the number
        of digits to compare.

        When comparing two numbers, if the first number has magnitude less
        than 1e-5, we compare the two numbers directly and check whether
        they are equivalent within the specified precision. Otherwise, we
        compare the **ratio** of the second number to the first number and
        check whether it is equivalent to 1 within the specified precision.

        .. deprecated:: 1.1.0
           Use `rtol` and `atol` instead to define relative/absolute
           tolerance, respectively. Similar to :func:`math.isclose`.
    rtol : float, default 1e-5
        Relative tolerance.

        .. versionadded:: 1.1.0
    atol : float, default 1e-8
        Absolute tolerance.

        .. versionadded:: 1.1.0
    """
    ...

def assert_dict_equal(left, right, compare_keys: bool = ...) -> None:
    ...

def assert_index_equal(left: Index, right: Index, exact: bool | str = ..., check_names: bool = ..., check_less_precise: bool | int | NoDefault = ..., check_exact: bool = ..., check_categorical: bool = ..., check_order: bool = ..., rtol: float = ..., atol: float = ..., obj: str = ...) -> None:
    """
    Check that left and right Index are equal.

    Parameters
    ----------
    left : Index
    right : Index
    exact : bool or {'equiv'}, default 'equiv'
        Whether to check the Index class, dtype and inferred_type
        are identical. If 'equiv', then RangeIndex can be substituted for
        Int64Index as well.
    check_names : bool, default True
        Whether to check the names attribute.
    check_less_precise : bool or int, default False
        Specify comparison precision. Only used when check_exact is False.
        5 digits (False) or 3 digits (True) after decimal points are compared.
        If int, then specify the digits to compare.

        .. deprecated:: 1.1.0
           Use `rtol` and `atol` instead to define relative/absolute
           tolerance, respectively. Similar to :func:`math.isclose`.
    check_exact : bool, default True
        Whether to compare number exactly.
    check_categorical : bool, default True
        Whether to compare internal Categorical exactly.
    check_order : bool, default True
        Whether to compare the order of index entries as well as their values.
        If True, both indexes must contain the same elements, in the same order.
        If False, both indexes must contain the same elements, but in any order.

        .. versionadded:: 1.2.0
    rtol : float, default 1e-5
        Relative tolerance. Only used when check_exact is False.

        .. versionadded:: 1.1.0
    atol : float, default 1e-8
        Absolute tolerance. Only used when check_exact is False.

        .. versionadded:: 1.1.0
    obj : str, default 'Index'
        Specify object name being compared, internally used to show appropriate
        assertion message.

    Examples
    --------
    >>> from pandas import testing as tm
    >>> a = pd.Index([1, 2, 3])
    >>> b = pd.Index([1, 2, 3])
    >>> tm.assert_index_equal(a, b)
    """
    ...

def assert_class_equal(left, right, exact: bool | str = ..., obj=...) -> None:
    """
    Checks classes are equal.
    """
    ...

def assert_attr_equal(attr: str, left, right, obj: str = ...) -> None:
    """
    Check attributes are equal. Both objects must have attribute.

    Parameters
    ----------
    attr : str
        Attribute name being compared.
    left : object
    right : object
    obj : str, default 'Attributes'
        Specify object name being compared, internally used to show appropriate
        assertion message
    """
    ...

def assert_is_valid_plot_return_object(objs) -> None:
    ...

def assert_is_sorted(seq) -> None:
    """Assert that the sequence is sorted."""
    ...

def assert_categorical_equal(left, right, check_dtype=..., check_category_order=..., obj=...) -> None:
    """
    Test that Categoricals are equivalent.

    Parameters
    ----------
    left : Categorical
    right : Categorical
    check_dtype : bool, default True
        Check that integer dtype of the codes are the same.
    check_category_order : bool, default True
        Whether the order of the categories should be compared, which
        implies identical integer codes.  If False, only the resulting
        values are compared.  The ordered attribute is
        checked regardless.
    obj : str, default 'Categorical'
        Specify object name being compared, internally used to show appropriate
        assertion message.
    """
    ...

def assert_interval_array_equal(left, right, exact=..., obj=...) -> None:
    """
    Test that two IntervalArrays are equivalent.

    Parameters
    ----------
    left, right : IntervalArray
        The IntervalArrays to compare.
    exact : bool or {'equiv'}, default 'equiv'
        Whether to check the Index class, dtype and inferred_type
        are identical. If 'equiv', then RangeIndex can be substituted for
        Int64Index as well.
    obj : str, default 'IntervalArray'
        Specify object name being compared, internally used to show appropriate
        assertion message
    """
    ...

def assert_period_array_equal(left, right, obj=...) -> None:
    ...

def assert_datetime_array_equal(left, right, obj=..., check_freq=...) -> None:
    ...

def assert_timedelta_array_equal(left, right, obj=..., check_freq=...) -> None:
    ...

def raise_assert_detail(obj, message, left, right, diff=..., index_values=...):
    ...

def assert_numpy_array_equal(left, right, strict_nan=..., check_dtype: bool | Literal["equiv"] = ..., err_msg=..., check_same=..., obj=..., index_values=...) -> None:
    """
    Check that 'np.ndarray' is equivalent.

    Parameters
    ----------
    left, right : numpy.ndarray or iterable
        The two arrays to be compared.
    strict_nan : bool, default False
        If True, consider NaN and None to be different.
    check_dtype : bool, default True
        Check dtype if both a and b are np.ndarray.
    err_msg : str, default None
        If provided, used as assertion message.
    check_same : None|'copy'|'same', default None
        Ensure left and right refer/do not refer to the same memory area.
    obj : str, default 'numpy array'
        Specify object name being compared, internally used to show appropriate
        assertion message.
    index_values : numpy.ndarray, default None
        optional index (shared by both left and right), used in output.
    """
    ...

def assert_extension_array_equal(left, right, check_dtype: bool | Literal["equiv"] = ..., index_values=..., check_less_precise=..., check_exact=..., rtol: float = ..., atol: float = ...) -> None:
    """
    Check that left and right ExtensionArrays are equal.

    Parameters
    ----------
    left, right : ExtensionArray
        The two arrays to compare.
    check_dtype : bool, default True
        Whether to check if the ExtensionArray dtypes are identical.
    index_values : numpy.ndarray, default None
        Optional index (shared by both left and right), used in output.
    check_less_precise : bool or int, default False
        Specify comparison precision. Only used when check_exact is False.
        5 digits (False) or 3 digits (True) after decimal points are compared.
        If int, then specify the digits to compare.

        .. deprecated:: 1.1.0
           Use `rtol` and `atol` instead to define relative/absolute
           tolerance, respectively. Similar to :func:`math.isclose`.
    check_exact : bool, default False
        Whether to compare number exactly.
    rtol : float, default 1e-5
        Relative tolerance. Only used when check_exact is False.

        .. versionadded:: 1.1.0
    atol : float, default 1e-8
        Absolute tolerance. Only used when check_exact is False.

        .. versionadded:: 1.1.0

    Notes
    -----
    Missing values are checked separately from valid values.
    A mask of missing values is computed for each and checked to match.
    The remaining all-valid values are cast to object dtype and checked.

    Examples
    --------
    >>> from pandas import testing as tm
    >>> a = pd.Series([1, 2, 3, 4])
    >>> b, c = a.array, a.array
    >>> tm.assert_extension_array_equal(b, c)
    """
    ...

def assert_series_equal(left, right, check_dtype: bool | Literal["equiv"] = ..., check_index_type: bool | Literal["equiv"] = ..., check_series_type=..., check_less_precise: bool | int | NoDefault = ..., check_names=..., check_exact=..., check_datetimelike_compat=..., check_categorical=..., check_category_order=..., check_freq=..., check_flags=..., rtol=..., atol=..., obj=..., *, check_index=..., check_like=...) -> None:
    """
    Check that left and right Series are equal.

    Parameters
    ----------
    left : Series
    right : Series
    check_dtype : bool, default True
        Whether to check the Series dtype is identical.
    check_index_type : bool or {'equiv'}, default 'equiv'
        Whether to check the Index class, dtype and inferred_type
        are identical.
    check_series_type : bool, default True
         Whether to check the Series class is identical.
    check_less_precise : bool or int, default False
        Specify comparison precision. Only used when check_exact is False.
        5 digits (False) or 3 digits (True) after decimal points are compared.
        If int, then specify the digits to compare.

        When comparing two numbers, if the first number has magnitude less
        than 1e-5, we compare the two numbers directly and check whether
        they are equivalent within the specified precision. Otherwise, we
        compare the **ratio** of the second number to the first number and
        check whether it is equivalent to 1 within the specified precision.

        .. deprecated:: 1.1.0
           Use `rtol` and `atol` instead to define relative/absolute
           tolerance, respectively. Similar to :func:`math.isclose`.
    check_names : bool, default True
        Whether to check the Series and Index names attribute.
    check_exact : bool, default False
        Whether to compare number exactly.
    check_datetimelike_compat : bool, default False
        Compare datetime-like which is comparable ignoring dtype.
    check_categorical : bool, default True
        Whether to compare internal Categorical exactly.
    check_category_order : bool, default True
        Whether to compare category order of internal Categoricals.

        .. versionadded:: 1.0.2
    check_freq : bool, default True
        Whether to check the `freq` attribute on a DatetimeIndex or TimedeltaIndex.

        .. versionadded:: 1.1.0
    check_flags : bool, default True
        Whether to check the `flags` attribute.

        .. versionadded:: 1.2.0

    rtol : float, default 1e-5
        Relative tolerance. Only used when check_exact is False.

        .. versionadded:: 1.1.0
    atol : float, default 1e-8
        Absolute tolerance. Only used when check_exact is False.

        .. versionadded:: 1.1.0
    obj : str, default 'Series'
        Specify object name being compared, internally used to show appropriate
        assertion message.
    check_index : bool, default True
        Whether to check index equivalence. If False, then compare only values.

        .. versionadded:: 1.3.0
    check_like : bool, default False
        If True, ignore the order of the index. Must be False if check_index is False.
        Note: same labels must be with the same data.

        .. versionadded:: 1.5.0

    Examples
    --------
    >>> from pandas import testing as tm
    >>> a = pd.Series([1, 2, 3, 4])
    >>> b = pd.Series([1, 2, 3, 4])
    >>> tm.assert_series_equal(a, b)
    """
    ...

def assert_frame_equal(left, right, check_dtype: bool | Literal["equiv"] = ..., check_index_type: bool | Literal["equiv"] = ..., check_column_type=..., check_frame_type=..., check_less_precise=..., check_names=..., by_blocks=..., check_exact=..., check_datetimelike_compat=..., check_categorical=..., check_like=..., check_freq=..., check_flags=..., rtol=..., atol=..., obj=...) -> None:
    """
    Check that left and right DataFrame are equal.

    This function is intended to compare two DataFrames and output any
    differences. It is mostly intended for use in unit tests.
    Additional parameters allow varying the strictness of the
    equality checks performed.

    Parameters
    ----------
    left : DataFrame
        First DataFrame to compare.
    right : DataFrame
        Second DataFrame to compare.
    check_dtype : bool, default True
        Whether to check the DataFrame dtype is identical.
    check_index_type : bool or {'equiv'}, default 'equiv'
        Whether to check the Index class, dtype and inferred_type
        are identical.
    check_column_type : bool or {'equiv'}, default 'equiv'
        Whether to check the columns class, dtype and inferred_type
        are identical. Is passed as the ``exact`` argument of
        :func:`assert_index_equal`.
    check_frame_type : bool, default True
        Whether to check the DataFrame class is identical.
    check_less_precise : bool or int, default False
        Specify comparison precision. Only used when check_exact is False.
        5 digits (False) or 3 digits (True) after decimal points are compared.
        If int, then specify the digits to compare.

        When comparing two numbers, if the first number has magnitude less
        than 1e-5, we compare the two numbers directly and check whether
        they are equivalent within the specified precision. Otherwise, we
        compare the **ratio** of the second number to the first number and
        check whether it is equivalent to 1 within the specified precision.

        .. deprecated:: 1.1.0
           Use `rtol` and `atol` instead to define relative/absolute
           tolerance, respectively. Similar to :func:`math.isclose`.
    check_names : bool, default True
        Whether to check that the `names` attribute for both the `index`
        and `column` attributes of the DataFrame is identical.
    by_blocks : bool, default False
        Specify how to compare internal data. If False, compare by columns.
        If True, compare by blocks.
    check_exact : bool, default False
        Whether to compare number exactly.
    check_datetimelike_compat : bool, default False
        Compare datetime-like which is comparable ignoring dtype.
    check_categorical : bool, default True
        Whether to compare internal Categorical exactly.
    check_like : bool, default False
        If True, ignore the order of index & columns.
        Note: index labels must match their respective rows
        (same as in columns) - same labels must be with the same data.
    check_freq : bool, default True
        Whether to check the `freq` attribute on a DatetimeIndex or TimedeltaIndex.

        .. versionadded:: 1.1.0
    check_flags : bool, default True
        Whether to check the `flags` attribute.
    rtol : float, default 1e-5
        Relative tolerance. Only used when check_exact is False.

        .. versionadded:: 1.1.0
    atol : float, default 1e-8
        Absolute tolerance. Only used when check_exact is False.

        .. versionadded:: 1.1.0
    obj : str, default 'DataFrame'
        Specify object name being compared, internally used to show appropriate
        assertion message.

    See Also
    --------
    assert_series_equal : Equivalent method for asserting Series equality.
    DataFrame.equals : Check DataFrame equality.

    Examples
    --------
    This example shows comparing two DataFrames that are equal
    but with columns of differing dtypes.

    >>> from pandas.testing import assert_frame_equal
    >>> df1 = pd.DataFrame({'a': [1, 2], 'b': [3, 4]})
    >>> df2 = pd.DataFrame({'a': [1, 2], 'b': [3.0, 4.0]})

    df1 equals itself.

    >>> assert_frame_equal(df1, df1)

    df1 differs from df2 as column 'b' is of a different type.

    >>> assert_frame_equal(df1, df2)
    Traceback (most recent call last):
    ...
    AssertionError: Attributes of DataFrame.iloc[:, 1] (column name="b") are different

    Attribute "dtype" are different
    [left]:  int64
    [right]: float64

    Ignore differing dtypes in columns with check_dtype.

    >>> assert_frame_equal(df1, df2, check_dtype=False)
    """
    ...

def assert_equal(left, right, **kwargs) -> None:
    """
    Wrapper for tm.assert_*_equal to dispatch to the appropriate test function.

    Parameters
    ----------
    left, right : Index, Series, DataFrame, ExtensionArray, or np.ndarray
        The two items to be compared.
    **kwargs
        All keyword arguments are passed through to the underlying assert method.
    """
    ...

def assert_sp_array_equal(left, right) -> None:
    """
    Check that the left and right SparseArray are equal.

    Parameters
    ----------
    left : SparseArray
    right : SparseArray
    """
    ...

def assert_contains_all(iterable, dic) -> None:
    ...

def assert_copy(iter1, iter2, **eql_kwargs) -> None:
    """
    iter1, iter2: iterables that produce elements
    comparable with assert_almost_equal

    Checks that the elements are equal, but not
    the same object. (Does not check that items
    in sequences are also not the same object)
    """
    ...

def is_extension_array_dtype_and_needs_i8_conversion(left_dtype, right_dtype) -> bool:
    """
    Checks that we have the combination of an ExtensionArraydtype and
    a dtype that should be converted to int64

    Returns
    -------
    bool

    Related to issue #37609
    """
    ...

def assert_indexing_slices_equivalent(ser: Series, l_slc: slice, i_slc: slice) -> None:
    """
    Check that ser.iloc[i_slc] matches ser.loc[l_slc] and, if applicable,
    ser[l_slc].
    """
    ...

def assert_metadata_equivalent(left, right) -> None:
    """
    Check that ._metadata attributes are equivalent.
    """
    ...




"""
This type stub file was generated by pyright.
"""

from typing import TYPE_CHECKING
from pandas._typing import DtypeObj, JSONSerializable
from pandas.core.dtypes.dtypes import CategoricalDtype
from pandas import DataFrame, Series

"""
Table Schema builders

https://specs.frictionlessdata.io/json-table-schema/
"""
if TYPE_CHECKING:
    ...
loads = ...
TABLE_SCHEMA_VERSION = ...
def as_json_table_type(x: DtypeObj) -> str:
    """
    Convert a NumPy / pandas type to its corresponding json_table.

    Parameters
    ----------
    x : np.dtype or ExtensionDtype

    Returns
    -------
    str
        the Table Schema data types

    Notes
    -----
    This table shows the relationship between NumPy / pandas dtypes,
    and Table Schema dtypes.

    ==============  =================
    Pandas type     Table Schema type
    ==============  =================
    int64           integer
    float64         number
    bool            boolean
    datetime64[ns]  datetime
    timedelta64[ns] duration
    object          str
    categorical     any
    =============== =================
    """
    ...

def set_default_names(data):
    """Sets index names to 'index' for regular, or 'level_x' for Multi"""
    ...

def convert_pandas_type_to_json_field(arr) -> dict[str, JSONSerializable]:
    ...

def convert_json_field_to_pandas_type(field) -> str | CategoricalDtype:
    """
    Converts a JSON field descriptor into its corresponding NumPy / pandas type

    Parameters
    ----------
    field
        A JSON field descriptor

    Returns
    -------
    dtype

    Raises
    ------
    ValueError
        If the type of the provided field is unknown or currently unsupported

    Examples
    --------
    >>> convert_json_field_to_pandas_type({"name": "an_int", "type": "integer"})
    'int64'

    >>> convert_json_field_to_pandas_type(
    ...     {
    ...         "name": "a_categorical",
    ...         "type": "any",
    ...         "constraints": {"enum": ["a", "b", "c"]},
    ...         "ordered": True,
    ...     }
    ... )
    CategoricalDtype(categories=['a', 'b', 'c'], ordered=True)

    >>> convert_json_field_to_pandas_type({"name": "a_datetime", "type": "datetime"})
    'datetime64[ns]'

    >>> convert_json_field_to_pandas_type(
    ...     {"name": "a_datetime_with_tz", "type": "datetime", "tz": "US/Central"}
    ... )
    'datetime64[ns, US/Central]'
    """
    ...

def build_table_schema(data: DataFrame | Series, index: bool = ..., primary_key: bool | None = ..., version: bool = ...) -> dict[str, JSONSerializable]:
    """
    Create a Table schema from ``data``.

    Parameters
    ----------
    data : Series, DataFrame
    index : bool, default True
        Whether to include ``data.index`` in the schema.
    primary_key : bool or None, default True
        Column names to designate as the primary key.
        The default `None` will set `'primaryKey'` to the index
        level or levels if the index is unique.
    version : bool, default True
        Whether to include a field `pandas_version` with the version
        of pandas that last revised the table schema. This version
        can be different from the installed pandas version.

    Returns
    -------
    schema : dict

    Notes
    -----
    See `Table Schema
    <https://pandas.pydata.org/docs/user_guide/io.html#table-schema>`__ for
    conversion types.
    Timedeltas as converted to ISO8601 duration format with
    9 decimal places after the seconds field for nanosecond precision.

    Categoricals are converted to the `any` dtype, and use the `enum` field
    constraint to list the allowed values. The `ordered` attribute is included
    in an `ordered` field.

    Examples
    --------
    >>> df = pd.DataFrame(
    ...     {'A': [1, 2, 3],
    ...      'B': ['a', 'b', 'c'],
    ...      'C': pd.date_range('2016-01-01', freq='d', periods=3),
    ...     }, index=pd.Index(range(3), name='idx'))
    >>> build_table_schema(df)
    {'fields': \
[{'name': 'idx', 'type': 'integer'}, \
{'name': 'A', 'type': 'integer'}, \
{'name': 'B', 'type': 'string'}, \
{'name': 'C', 'type': 'datetime'}], \
'primaryKey': ['idx'], \
'pandas_version': '1.4.0'}
    """
    ...

def parse_table_schema(json, precise_float):
    """
    Builds a DataFrame from a given schema

    Parameters
    ----------
    json :
        A JSON table schema
    precise_float : bool
        Flag controlling precision when decoding string to double values, as
        dictated by ``read_json``

    Returns
    -------
    df : DataFrame

    Raises
    ------
    NotImplementedError
        If the JSON table schema contains either timezone or timedelta data

    Notes
    -----
        Because :func:`DataFrame.to_json` uses the string 'index' to denote a
        name-less :class:`Index`, this function sets the name of the returned
        :class:`DataFrame` to ``None`` when said string is encountered with a
        normal :class:`Index`. For a :class:`MultiIndex`, the same limitation
        applies to any strings beginning with 'level_'. Therefore, an
        :class:`Index` name of 'index'  and :class:`MultiIndex` names starting
        with 'level_' are not supported.

    See Also
    --------
    build_table_schema : Inverse function.
    pandas.read_json
    """
    ...




"""
This type stub file was generated by pyright.
"""

import numpy as np
from pandas.util.version import Version

""" support numpy compatibility across versions """
_np_version = ...
_nlv = ...
np_version_under1p21 = ...
np_version_under1p22 = ...
np_version_gte1p22 = ...
is_numpy_dev = ...
_min_numpy_ver = ...
if is_numpy_dev or notnp_version_under1p22:
    np_percentile_argname = ...
else:
    np_percentile_argname = ...
if _nlv < Version(_min_numpy_ver):
    ...
__all__ = ["np", "_np_version", "is_numpy_dev"]



"""
This type stub file was generated by pyright.
"""

from typing import Final

magic: Final = ...
align_1_checker_value: Final = ...
align_1_offset: Final = ...
align_1_length: Final = ...
align_1_value: Final = ...
u64_byte_checker_value: Final = ...
align_2_offset: Final = ...
align_2_length: Final = ...
align_2_value: Final = ...
endianness_offset: Final = ...
endianness_length: Final = ...
platform_offset: Final = ...
platform_length: Final = ...
encoding_offset: Final = ...
encoding_length: Final = ...
dataset_offset: Final = ...
dataset_length: Final = ...
file_type_offset: Final = ...
file_type_length: Final = ...
date_created_offset: Final = ...
date_created_length: Final = ...
date_modified_offset: Final = ...
date_modified_length: Final = ...
header_size_offset: Final = ...
header_size_length: Final = ...
page_size_offset: Final = ...
page_size_length: Final = ...
page_count_offset: Final = ...
page_count_length: Final = ...
sas_release_offset: Final = ...
sas_release_length: Final = ...
sas_server_type_offset: Final = ...
sas_server_type_length: Final = ...
os_version_number_offset: Final = ...
os_version_number_length: Final = ...
os_maker_offset: Final = ...
os_maker_length: Final = ...
os_name_offset: Final = ...
os_name_length: Final = ...
page_bit_offset_x86: Final = ...
page_bit_offset_x64: Final = ...
subheader_pointer_length_x86: Final = ...
subheader_pointer_length_x64: Final = ...
page_type_offset: Final = ...
page_type_length: Final = ...
block_count_offset: Final = ...
block_count_length: Final = ...
subheader_count_offset: Final = ...
subheader_count_length: Final = ...
page_type_mask: Final = ...
page_type_mask2: Final = ...
page_meta_type: Final = ...
page_data_type: Final = ...
page_mix_type: Final = ...
page_amd_type: Final = ...
page_meta2_type: Final = ...
page_comp_type: Final = ...
page_meta_types: Final = ...
subheader_pointers_offset: Final = ...
truncated_subheader_id: Final = ...
compressed_subheader_id: Final = ...
compressed_subheader_type: Final = ...
text_block_size_length: Final = ...
row_length_offset_multiplier: Final = ...
row_count_offset_multiplier: Final = ...
col_count_p1_multiplier: Final = ...
col_count_p2_multiplier: Final = ...
row_count_on_mix_page_offset_multiplier: Final = ...
column_name_pointer_length: Final = ...
column_name_text_subheader_offset: Final = ...
column_name_text_subheader_length: Final = ...
column_name_offset_offset: Final = ...
column_name_offset_length: Final = ...
column_name_length_offset: Final = ...
column_name_length_length: Final = ...
column_data_offset_offset: Final = ...
column_data_length_offset: Final = ...
column_data_length_length: Final = ...
column_type_offset: Final = ...
column_type_length: Final = ...
column_format_text_subheader_index_offset: Final = ...
column_format_text_subheader_index_length: Final = ...
column_format_offset_offset: Final = ...
column_format_offset_length: Final = ...
column_format_length_offset: Final = ...
column_format_length_length: Final = ...
column_label_text_subheader_index_offset: Final = ...
column_label_text_subheader_index_length: Final = ...
column_label_offset_offset: Final = ...
column_label_offset_length: Final = ...
column_label_length_offset: Final = ...
column_label_length_length: Final = ...
rle_compression: Final = ...
rdc_compression: Final = ...
compression_literals: Final = ...
encoding_names: Final = ...
class SASIndex:
    row_size_index: Final = ...
    column_size_index: Final = ...
    subheader_counts_index: Final = ...
    column_text_index: Final = ...
    column_name_index: Final = ...
    column_attributes_index: Final = ...
    format_and_label_index: Final = ...
    column_list_index: Final = ...
    data_subheader_index: Final = ...


subheader_signature_to_index: Final = ...
sas_date_formats: Final = ...
sas_datetime_formats: Final = ...



"""
This type stub file was generated by pyright.
"""

import abc
from typing import Any, Callable, Hashable, Iterable, Literal, Mapping, Sequence, overload
from pandas._typing import DtypeArg, FilePath, IntStrT, ReadBuffer, StorageOptions, WriteExcelBuffer
from pandas.util._decorators import Appender, deprecate_kwarg, deprecate_nonkeyword_arguments, doc
from pandas.core.frame import DataFrame
from pandas.core.shared_docs import _shared_docs
from pandas.io.common import IOHandles

_read_excel_doc = ...
@overload
def read_excel(io, sheet_name: str | int = ..., header: int | Sequence[int] | None = ..., names: list[str] | None = ..., index_col: int | Sequence[int] | None = ..., usecols: int | str | Sequence[int] | Sequence[str] | Callable[[str], bool] | None = ..., squeeze: bool | None = ..., dtype: DtypeArg | None = ..., engine: Literal["xlrd", "openpyxl", "odf", "pyxlsb"] | None = ..., converters: dict[str, Callable] | dict[int, Callable] | None = ..., true_values: Iterable[Hashable] | None = ..., false_values: Iterable[Hashable] | None = ..., skiprows: Sequence[int] | int | Callable[[int], object] | None = ..., nrows: int | None = ..., na_values=..., keep_default_na: bool = ..., na_filter: bool = ..., verbose: bool = ..., parse_dates: list | dict | bool = ..., date_parser: Callable | None = ..., thousands: str | None = ..., decimal: str = ..., comment: str | None = ..., skipfooter: int = ..., convert_float: bool | None = ..., mangle_dupe_cols: bool = ..., storage_options: StorageOptions = ...) -> DataFrame:
    ...

@overload
def read_excel(io, sheet_name: list[IntStrT] | None, header: int | Sequence[int] | None = ..., names: list[str] | None = ..., index_col: int | Sequence[int] | None = ..., usecols: int | str | Sequence[int] | Sequence[str] | Callable[[str], bool] | None = ..., squeeze: bool | None = ..., dtype: DtypeArg | None = ..., engine: Literal["xlrd", "openpyxl", "odf", "pyxlsb"] | None = ..., converters: dict[str, Callable] | dict[int, Callable] | None = ..., true_values: Iterable[Hashable] | None = ..., false_values: Iterable[Hashable] | None = ..., skiprows: Sequence[int] | int | Callable[[int], object] | None = ..., nrows: int | None = ..., na_values=..., keep_default_na: bool = ..., na_filter: bool = ..., verbose: bool = ..., parse_dates: list | dict | bool = ..., date_parser: Callable | None = ..., thousands: str | None = ..., decimal: str = ..., comment: str | None = ..., skipfooter: int = ..., convert_float: bool | None = ..., mangle_dupe_cols: bool = ..., storage_options: StorageOptions = ...) -> dict[IntStrT, DataFrame]:
    ...

@doc(storage_options=_shared_docs["storage_options"])
@deprecate_kwarg(old_arg_name="mangle_dupe_cols", new_arg_name=None)
@deprecate_nonkeyword_arguments(allowed_args=["io", "sheet_name"], version="2.0")
@Appender(_read_excel_doc)
def read_excel(io, sheet_name: str | int | list[IntStrT] | None = ..., header: int | Sequence[int] | None = ..., names: list[str] | None = ..., index_col: int | Sequence[int] | None = ..., usecols: int | str | Sequence[int] | Sequence[str] | Callable[[str], bool] | None = ..., squeeze: bool | None = ..., dtype: DtypeArg | None = ..., engine: Literal["xlrd", "openpyxl", "odf", "pyxlsb"] | None = ..., converters: dict[str, Callable] | dict[int, Callable] | None = ..., true_values: Iterable[Hashable] | None = ..., false_values: Iterable[Hashable] | None = ..., skiprows: Sequence[int] | int | Callable[[int], object] | None = ..., nrows: int | None = ..., na_values=..., keep_default_na: bool = ..., na_filter: bool = ..., verbose: bool = ..., parse_dates: list | dict | bool = ..., date_parser: Callable | None = ..., thousands: str | None = ..., decimal: str = ..., comment: str | None = ..., skipfooter: int = ..., convert_float: bool | None = ..., mangle_dupe_cols: bool = ..., storage_options: StorageOptions = ...) -> DataFrame | dict[IntStrT, DataFrame]:
    ...

class BaseExcelReader(metaclass=abc.ABCMeta):
    def __init__(self, filepath_or_buffer, storage_options: StorageOptions = ...) -> None:
        ...
    
    @abc.abstractmethod
    def load_workbook(self, filepath_or_buffer): # -> None:
        ...
    
    def close(self) -> None:
        ...
    
    @property
    @abc.abstractmethod
    def sheet_names(self) -> list[str]:
        ...
    
    @abc.abstractmethod
    def get_sheet_by_name(self, name: str): # -> None:
        ...
    
    @abc.abstractmethod
    def get_sheet_by_index(self, index: int): # -> None:
        ...
    
    @abc.abstractmethod
    def get_sheet_data(self, sheet, convert_float: bool, rows: int | None = ...): # -> None:
        ...
    
    def raise_if_bad_sheet_by_index(self, index: int) -> None:
        ...
    
    def raise_if_bad_sheet_by_name(self, name: str) -> None:
        ...
    
    def parse(self, sheet_name: str | int | list[int] | list[str] | None = ..., header: int | Sequence[int] | None = ..., names=..., index_col: int | Sequence[int] | None = ..., usecols=..., squeeze: bool | None = ..., dtype: DtypeArg | None = ..., true_values: Iterable[Hashable] | None = ..., false_values: Iterable[Hashable] | None = ..., skiprows: Sequence[int] | int | Callable[[int], object] | None = ..., nrows: int | None = ..., na_values=..., verbose: bool = ..., parse_dates: list | dict | bool = ..., date_parser: Callable | None = ..., thousands: str | None = ..., decimal: str = ..., comment: str | None = ..., skipfooter: int = ..., convert_float: bool | None = ..., mangle_dupe_cols: bool = ..., **kwds): # -> dict[Unknown, Unknown]:
        ...
    


@doc(storage_options=_shared_docs["storage_options"])
class ExcelWriter(metaclass=abc.ABCMeta):
    """
    Class for writing DataFrame objects into excel sheets.

    Default is to use:

    * `xlwt <https://pypi.org/project/xlwt/>`__ for xls files
    * `xlsxwriter <https://pypi.org/project/XlsxWriter/>`__ for xlsx files if xlsxwriter
      is installed otherwise `openpyxl <https://pypi.org/project/openpyxl/>`__
    * `odswriter <https://pypi.org/project/odswriter/>`__ for ods files

    See ``DataFrame.to_excel`` for typical usage.

    The writer should be used as a context manager. Otherwise, call `close()` to save
    and close any opened file handles.

    Parameters
    ----------
    path : str or typing.BinaryIO
        Path to xls or xlsx or ods file.
    engine : str (optional)
        Engine to use for writing. If None, defaults to
        ``io.excel.<extension>.writer``.  NOTE: can only be passed as a keyword
        argument.

        .. deprecated:: 1.2.0

            As the `xlwt <https://pypi.org/project/xlwt/>`__ package is no longer
            maintained, the ``xlwt`` engine will be removed in a future
            version of pandas.

    date_format : str, default None
        Format string for dates written into Excel files (e.g. 'YYYY-MM-DD').
    datetime_format : str, default None
        Format string for datetime objects written into Excel files.
        (e.g. 'YYYY-MM-DD HH:MM:SS').
    mode : {{'w', 'a'}}, default 'w'
        File mode to use (write or append). Append does not work with fsspec URLs.
    {storage_options}

        .. versionadded:: 1.2.0

    if_sheet_exists : {{'error', 'new', 'replace', 'overlay'}}, default 'error'
        How to behave when trying to write to a sheet that already
        exists (append mode only).

        * error: raise a ValueError.
        * new: Create a new sheet, with a name determined by the engine.
        * replace: Delete the contents of the sheet before writing to it.
        * overlay: Write contents to the existing sheet without removing the old
          contents.

        .. versionadded:: 1.3.0

        .. versionchanged:: 1.4.0

           Added ``overlay`` option

    engine_kwargs : dict, optional
        Keyword arguments to be passed into the engine. These will be passed to
        the following functions of the respective engines:

        * xlsxwriter: ``xlsxwriter.Workbook(file, **engine_kwargs)``
        * openpyxl (write mode): ``openpyxl.Workbook(**engine_kwargs)``
        * openpyxl (append mode): ``openpyxl.load_workbook(file, **engine_kwargs)``
        * odswriter: ``odf.opendocument.OpenDocumentSpreadsheet(**engine_kwargs)``

        .. versionadded:: 1.3.0
    **kwargs : dict, optional
        Keyword arguments to be passed into the engine.

        .. deprecated:: 1.3.0

            Use engine_kwargs instead.

    Notes
    -----
    For compatibility with CSV writers, ExcelWriter serializes lists
    and dicts to strings before writing.

    Examples
    --------
    Default usage:

    >>> df = pd.DataFrame([["ABC", "XYZ"]], columns=["Foo", "Bar"])  # doctest: +SKIP
    >>> with pd.ExcelWriter("path_to_file.xlsx") as writer:
    ...     df.to_excel(writer)  # doctest: +SKIP

    To write to separate sheets in a single file:

    >>> df1 = pd.DataFrame([["AAA", "BBB"]], columns=["Spam", "Egg"])  # doctest: +SKIP
    >>> df2 = pd.DataFrame([["ABC", "XYZ"]], columns=["Foo", "Bar"])  # doctest: +SKIP
    >>> with pd.ExcelWriter("path_to_file.xlsx") as writer:
    ...     df1.to_excel(writer, sheet_name="Sheet1")  # doctest: +SKIP
    ...     df2.to_excel(writer, sheet_name="Sheet2")  # doctest: +SKIP

    You can set the date format or datetime format:

    >>> from datetime import date, datetime  # doctest: +SKIP
    >>> df = pd.DataFrame(
    ...     [
    ...         [date(2014, 1, 31), date(1999, 9, 24)],
    ...         [datetime(1998, 5, 26, 23, 33, 4), datetime(2014, 2, 28, 13, 5, 13)],
    ...     ],
    ...     index=["Date", "Datetime"],
    ...     columns=["X", "Y"],
    ... )  # doctest: +SKIP
    >>> with pd.ExcelWriter(
    ...     "path_to_file.xlsx",
    ...     date_format="YYYY-MM-DD",
    ...     datetime_format="YYYY-MM-DD HH:MM:SS"
    ... ) as writer:
    ...     df.to_excel(writer)  # doctest: +SKIP

    You can also append to an existing Excel file:

    >>> with pd.ExcelWriter("path_to_file.xlsx", mode="a", engine="openpyxl") as writer:
    ...     df.to_excel(writer, sheet_name="Sheet3")  # doctest: +SKIP

    Here, the `if_sheet_exists` parameter can be set to replace a sheet if it
    already exists:

    >>> with ExcelWriter(
    ...     "path_to_file.xlsx",
    ...     mode="a",
    ...     engine="openpyxl",
    ...     if_sheet_exists="replace",
    ... ) as writer:
    ...     df.to_excel(writer, sheet_name="Sheet1")  # doctest: +SKIP

    You can also write multiple DataFrames to a single sheet. Note that the
    ``if_sheet_exists`` parameter needs to be set to ``overlay``:

    >>> with ExcelWriter("path_to_file.xlsx",
    ...     mode="a",
    ...     engine="openpyxl",
    ...     if_sheet_exists="overlay",
    ... ) as writer:
    ...     df1.to_excel(writer, sheet_name="Sheet1")
    ...     df2.to_excel(writer, sheet_name="Sheet1", startcol=3)  # doctest: +SKIP

    You can store Excel file in RAM:

    >>> import io
    >>> df = pd.DataFrame([["ABC", "XYZ"]], columns=["Foo", "Bar"])
    >>> buffer = io.BytesIO()
    >>> with pd.ExcelWriter(buffer) as writer:
    ...     df.to_excel(writer)

    You can pack Excel file into zip archive:

    >>> import zipfile  # doctest: +SKIP
    >>> df = pd.DataFrame([["ABC", "XYZ"]], columns=["Foo", "Bar"])  # doctest: +SKIP
    >>> with zipfile.ZipFile("path_to_file.zip", "w") as zf:
    ...     with zf.open("filename.xlsx", "w") as buffer:
    ...         with pd.ExcelWriter(buffer) as writer:
    ...             df.to_excel(writer)  # doctest: +SKIP

    You can specify additional arguments to the underlying engine:

    >>> with pd.ExcelWriter(
    ...     "path_to_file.xlsx",
    ...     engine="xlsxwriter",
    ...     engine_kwargs={{"options": {{"nan_inf_to_errors": True}}}}
    ... ) as writer:
    ...     df.to_excel(writer)  # doctest: +SKIP

    In append mode, ``engine_kwargs`` are passed through to
    openpyxl's ``load_workbook``:

    >>> with pd.ExcelWriter(
    ...     "path_to_file.xlsx",
    ...     engine="openpyxl",
    ...     mode="a",
    ...     engine_kwargs={{"keep_vba": True}}
    ... ) as writer:
    ...     df.to_excel(writer, sheet_name="Sheet2")  # doctest: +SKIP
    """
    _engine: str
    _supported_extensions: tuple[str, ...]
    def __new__(cls: type[ExcelWriter], path: FilePath | WriteExcelBuffer | ExcelWriter, engine: str | None = ..., date_format: str | None = ..., datetime_format: str | None = ..., mode: str = ..., storage_options: StorageOptions = ..., if_sheet_exists: Literal["error", "new", "replace", "overlay"] | None = ..., engine_kwargs: dict | None = ..., **kwargs) -> ExcelWriter:
        ...
    
    _path = ...
    @property
    def supported_extensions(self) -> tuple[str, ...]:
        """Extensions that writer engine supports."""
        ...
    
    @property
    def engine(self) -> str:
        """Name of engine."""
        ...
    
    @property
    @abc.abstractmethod
    def sheets(self) -> dict[str, Any]:
        """Mapping of sheet names to sheet objects."""
        ...
    
    @property
    @abc.abstractmethod
    def book(self): # -> None:
        """
        Book instance. Class type will depend on the engine used.

        This attribute can be used to access engine-specific features.
        """
        ...
    
    def write_cells(self, cells, sheet_name: str | None = ..., startrow: int = ..., startcol: int = ..., freeze_panes: tuple[int, int] | None = ...) -> None:
        """
        Write given formatted cells into Excel an excel sheet

        .. deprecated:: 1.5.0

        Parameters
        ----------
        cells : generator
            cell of formatted data to save to Excel sheet
        sheet_name : str, default None
            Name of Excel sheet, if None, then use self.cur_sheet
        startrow : upper left cell row to dump data frame
        startcol : upper left cell column to dump data frame
        freeze_panes: int tuple of length 2
            contains the bottom-most row and right-most column to freeze
        """
        ...
    
    def save(self) -> None:
        """
        Save workbook to disk.

        .. deprecated:: 1.5.0
        """
        ...
    
    def __init__(self, path: FilePath | WriteExcelBuffer | ExcelWriter, engine: str | None = ..., date_format: str | None = ..., datetime_format: str | None = ..., mode: str = ..., storage_options: StorageOptions = ..., if_sheet_exists: str | None = ..., engine_kwargs: dict[str, Any] | None = ..., **kwargs) -> None:
        ...
    
    @property
    def date_format(self) -> str:
        """
        Format string for dates written into Excel files (e.g. ‘YYYY-MM-DD’).
        """
        ...
    
    @property
    def datetime_format(self) -> str:
        """
        Format string for dates written into Excel files (e.g. ‘YYYY-MM-DD’).
        """
        ...
    
    @property
    def if_sheet_exists(self) -> str:
        """
        How to behave when writing to a sheet that already exists in append mode.
        """
        ...
    
    @property
    def cur_sheet(self): # -> None:
        """
        Current sheet for writing.

        .. deprecated:: 1.5.0
        """
        ...
    
    @property
    def handles(self) -> IOHandles[bytes]:
        """
        Handles to Excel sheets.

        .. deprecated:: 1.5.0
        """
        ...
    
    @property
    def path(self): # -> None:
        """
        Path to Excel file.

        .. deprecated:: 1.5.0
        """
        ...
    
    def __fspath__(self) -> str:
        ...
    
    @classmethod
    def check_extension(cls, ext: str) -> Literal[True]:
        """
        checks that path's extension against the Writer's supported
        extensions.  If it isn't supported, raises UnsupportedFiletypeError.
        """
        ...
    
    def __enter__(self) -> ExcelWriter:
        ...
    
    def __exit__(self, exc_type, exc_value, traceback) -> None:
        ...
    
    def close(self) -> None:
        """synonym for save, to make it more file-like"""
        ...
    


XLS_SIGNATURES = ...
ZIP_SIGNATURE = ...
PEEK_SIZE = ...
@doc(storage_options=_shared_docs["storage_options"])
def inspect_excel_format(content_or_path: FilePath | ReadBuffer[bytes], storage_options: StorageOptions = ...) -> str | None:
    """
    Inspect the path or content of an excel file and get its format.

    Adopted from xlrd: https://github.com/python-excel/xlrd.

    Parameters
    ----------
    content_or_path : str or file-like object
        Path to file or content of file to inspect. May be a URL.
    {storage_options}

    Returns
    -------
    str or None
        Format of file if it can be determined.

    Raises
    ------
    ValueError
        If resulting stream is empty.
    BadZipFile
        If resulting stream does not have an XLS signature and is not a valid zipfile.
    """
    ...

class ExcelFile:
    """
    Class for parsing tabular excel sheets into DataFrame objects.

    See read_excel for more documentation.

    Parameters
    ----------
    path_or_buffer : str, bytes, path object (pathlib.Path or py._path.local.LocalPath),
        a file-like object, xlrd workbook or openpyxl workbook.
        If a string or path object, expected to be a path to a
        .xls, .xlsx, .xlsb, .xlsm, .odf, .ods, or .odt file.
    engine : str, default None
        If io is not a buffer or path, this must be set to identify io.
        Supported engines: ``xlrd``, ``openpyxl``, ``odf``, ``pyxlsb``
        Engine compatibility :

        - ``xlrd`` supports old-style Excel files (.xls).
        - ``openpyxl`` supports newer Excel file formats.
        - ``odf`` supports OpenDocument file formats (.odf, .ods, .odt).
        - ``pyxlsb`` supports Binary Excel files.

        .. versionchanged:: 1.2.0

           The engine `xlrd <https://xlrd.readthedocs.io/en/latest/>`_
           now only supports old-style ``.xls`` files.
           When ``engine=None``, the following logic will be
           used to determine the engine:

           - If ``path_or_buffer`` is an OpenDocument format (.odf, .ods, .odt),
             then `odf <https://pypi.org/project/odfpy/>`_ will be used.
           - Otherwise if ``path_or_buffer`` is an xls format,
             ``xlrd`` will be used.
           - Otherwise if ``path_or_buffer`` is in xlsb format,
             `pyxlsb <https://pypi.org/project/pyxlsb/>`_ will be used.

           .. versionadded:: 1.3.0
           - Otherwise if `openpyxl <https://pypi.org/project/openpyxl/>`_ is installed,
             then ``openpyxl`` will be used.
           - Otherwise if ``xlrd >= 2.0`` is installed, a ``ValueError`` will be raised.
           - Otherwise ``xlrd`` will be used and a ``FutureWarning`` will be raised.
             This case will raise a ``ValueError`` in a future version of pandas.

           .. warning::

            Please do not report issues when using ``xlrd`` to read ``.xlsx`` files.
            This is not supported, switch to using ``openpyxl`` instead.
    """
    _engines: Mapping[str, Any] = ...
    def __init__(self, path_or_buffer, engine: str | None = ..., storage_options: StorageOptions = ...) -> None:
        ...
    
    def __fspath__(self): # -> BytesIO:
        ...
    
    def parse(self, sheet_name: str | int | list[int] | list[str] | None = ..., header: int | Sequence[int] | None = ..., names=..., index_col: int | Sequence[int] | None = ..., usecols=..., squeeze: bool | None = ..., converters=..., true_values: Iterable[Hashable] | None = ..., false_values: Iterable[Hashable] | None = ..., skiprows: Sequence[int] | int | Callable[[int], object] | None = ..., nrows: int | None = ..., na_values=..., parse_dates: list | dict | bool = ..., date_parser: Callable | None = ..., thousands: str | None = ..., comment: str | None = ..., skipfooter: int = ..., convert_float: bool | None = ..., mangle_dupe_cols: bool = ..., **kwds) -> DataFrame | dict[str, DataFrame] | dict[int, DataFrame]:
        """
        Parse specified sheet(s) into a DataFrame.

        Equivalent to read_excel(ExcelFile, ...)  See the read_excel
        docstring for more info on accepted parameters.

        Returns
        -------
        DataFrame or dict of DataFrames
            DataFrame from the passed in Excel file.
        """
        ...
    
    @property
    def book(self): # -> Any:
        ...
    
    @property
    def sheet_names(self): # -> Any:
        ...
    
    def close(self) -> None:
        """close io if necessary"""
        ...
    
    def __enter__(self) -> ExcelFile:
        ...
    
    def __exit__(self, exc_type, exc_value, traceback) -> None:
        ...
    
    def __del__(self) -> None:
        ...
    





# BoARIO : The Adaptative Regional Input Output model in python.
# Copyright (C) 2022  Samuel Juhel
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os
import sys
module_path = os.path.abspath(os.path.join('../'))
if module_path not in sys.path:
    sys.path.append(module_path)

module_path = os.path.abspath(os.path.join('./'))
if module_path not in sys.path:
    sys.path.append(module_path)

from boario.indicators import Indicators
import pathlib
import argparse
import logging
import re

parser = argparse.ArgumentParser(description="Produce indicators from one run folder")
parser.add_argument('folder', type=str, help='The str path to the folder')
logFormatter = logging.Formatter("%(asctime)s [%(levelname)-5.5s] %(name)s %(message)s", datefmt="%H:%M:%S")
scriptLogger = logging.getLogger("indicators_from_folder")
consoleHandler = logging.StreamHandler()
consoleHandler.setFormatter(logFormatter)

scriptLogger.addHandler(consoleHandler)
scriptLogger.setLevel(logging.INFO)
scriptLogger.propagate = False

def produce_indicator(folder):
    regex = re.compile(r"(?P<region>[A-Z]{2})_type_(?P<type>rebuilding|recover)_qdmg_(?:raw_|int_)?(?P<gdp_dmg>(?:[\d_]+%?)|(?:max|min))_Psi_(?P<psi>(?:0_\d+)|1_0)_inv_tau_(?P<inv_tau>\d+)(?:_inv_time_(?:\d+))?")
    m = regex.match(folder.name)
    if m is None:
        scriptLogger.warning("Directory {} didn't match regex".format(folder.name))
    else:
        indic = Indicators.from_folder(folder, folder/"indexes.json")
        indic.update_indicators()
        indic.indicators.copy()
        indic.write_indicators()

if __name__ == '__main__':
    scriptLogger.info('Starting Script')
    args = parser.parse_args()
    folder = pathlib.Path(args.folder).resolve()
    produce_indicator(folder)



# BoARIO : The Adaptative Regional Input Output model in python.
# Copyright (C) 2022  Samuel Juhel
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations
import pathlib
import numpy as np
from boario import logger
from boario.model_base import *
from boario.event import *
from pymrio.core.mriosystem import IOSystem

__all__ = ['ARIOModelPsi']
class ARIOModelPsi(ARIOBaseModel):
    """ An ARIO3 model with some additional features

    Added feature are parameter psi of production adjustment inventories constraint threshold, as well as a characteristic time of inventories resupplying and the alternative order module from Guan2020.

    Attributes
    ----------

    psi : float
          Value of the psi parameter. (see :ref:`boario-math`).
    restoration_tau : numpy.ndarray of int
                      Array of size `n_sector` setting for each inputs its characteristic restoration time in `temporal_units_by_step`. (see :ref:`boario-math`).
    Raises
    ------
    RuntimeError
        A RuntimeError can occur when data is inconsistent (negative stocks for
        instance)
    ValueError
    NotImplementedError
    """

    def __init__(self,
                 pym_mrio: IOSystem,
                 mrio_params: dict,
                 simulation_params: dict,
                 results_storage: pathlib.Path
                 ) -> None:

        super().__init__(pym_mrio, mrio_params, simulation_params, results_storage)
        logger.debug("Model is an ARIOModelPsi")

        self.psi = simulation_params['psi_param']
        inv = mrio_params['inventories_dict']
        inventories = [ np.inf if inv[k]=='inf' else inv[k] for k in sorted(inv.keys())]
        restoration_tau = [(self.n_temporal_units_by_step / simulation_params['inventory_restoration_tau']) if v >= INV_THRESHOLD else v for v in inventories] # for sector with no inventory TODO: reflect on that.
        self.restoration_tau = np.array(restoration_tau)
        #################################################################

    @property
    def inventory_constraints_opt(self) -> np.ndarray:
        return self.calc_inventory_constraints(self.production_opt)

    @property
    def inventory_constraints_act(self) -> np.ndarray:
        return self.calc_inventory_constraints(self.production)

    def calc_inventory_constraints(self, production: np.ndarray) -> np.ndarray:
        inventory_constraints = (np.tile(production, (self.n_sectors, 1)) * self.tech_mat) * self.psi
        np.multiply(inventory_constraints, np.tile(np.nan_to_num(self.inv_duration, posinf=0.)[:,np.newaxis],(1,self.n_regions*self.n_sectors)), out=inventory_constraints)
        return inventory_constraints

    def calc_production(self, current_temporal_unit:int) -> np.ndarray:
        r"""Compute and update actual production

        1. Compute ``production_opt`` and ``inventory_constraints`` as :

        .. math::
           :nowrap:

                \begin{alignat*}{4}
                      \iox^{\textrm{Opt}}(t) &= (x^{\textrm{Opt}}_{f}(t))_{f \in \firmsset} &&= \left ( \min \left ( d^{\textrm{Tot}}_{f}(t), x^{\textrm{Cap}}_{f}(t) \right ) \right )_{f \in \firmsset} && \text{Optimal production}\\
                      \mathbf{\ioinv}^{\textrm{Cons}}(t) &= (\omega^{\textrm{Cons},f}_p(t))_{\substack{p \in \sectorsset\\f \in \firmsset}} &&=
                    \begin{bmatrix}
                        \tau^{1}_1 & \hdots & \tau^{p}_1 \\
                        \vdots & \ddots & \vdots\\
                        \tau^1_n & \hdots & \tau^{p}_n
                    \end{bmatrix}
                    \odot \begin{bmatrix} \iox^{\textrm{Opt}}(t)\\ \vdots\\ \iox^{\textrm{Opt}}(t) \end{bmatrix} \odot \ioa^{\sectorsset} && \text{Inventory constraints} \\
                    &&&= \begin{bmatrix}
                        \tau^{1}_1 x^{\textrm{Opt}}_{1}(t) a_{11} & \hdots & \tau^{p}_1 x^{\textrm{Opt}}_{p}(t) a_{1p}\\
                        \vdots & \ddots & \vdots\\
                        \tau^1_n x^{\textrm{Opt}}_{1}(t) a_{n1} & \hdots & \tau^{p}_n x^{\textrm{Opt}}_{p}(t) a_{np}
                    \end{bmatrix} \cdot \psi && \\
                \end{alignat*}

        2. If stocks do not meet inventory_constraints for any inputs -> Decrease production accordingly :

        .. math::
           :nowrap:

                \begin{alignat*}{4}
                    \iox^{a}(t) &= (x^{a}_{f}(t))_{f \in \firmsset} &&= \left \{ \begin{aligned}
                                                           & x^{\textrm{Opt}}_{f}(t) & \text{if $\omega_{p}^f(t) \geq \omega^{\textrm{Cons},f}_p(t)$}\\
                                                           & x^{\textrm{Opt}}_{f}(t) \cdot \min_{p \in \sectorsset} \left ( \frac{\omega^s_{p}(t)}{\omega^{\textrm{Cons,f}}_p(t)} \right ) & \text{if $\omega_{p}^f(t) < \omega^{\textrm{Cons},f}_p(t)$}
                                                           \end{aligned} \right. \quad &&
                \end{alignat*}

        Also warns in log if such shortages happen.


        Parameters
        ----------
        current_temporal_unit : int
            current step number

        Returns
        -------
        np.ndarray
            A boolean NDArray `stock_constraint` of the same shape as `matrix_sock` (ie `(n_sectors,n_regions*n_sectors)`), with True for any input not meeting the inventory constraints.

        """
        #1.
        production_opt = self.production_opt.copy()
        inventory_constraints = self.inventory_constraints_opt.copy()
        # the following line is the same as in the ARIO base class except the multiplication by self.psi
        #2.
        if (stock_constraint := (self.matrix_stock < inventory_constraints) * self.matrix_share_thresh).any():
            if not self.in_shortage:
                logger.info('At least one industry entered shortage regime. (current temporal unit:{})'.format(current_temporal_unit))
            self.in_shortage = True
            self.had_shortage = True
            production_ratio_stock = np.ones(shape=self.matrix_stock.shape)
            np.divide(self.matrix_stock, inventory_constraints, out=production_ratio_stock, where=(self.matrix_share_thresh * (inventory_constraints!=0)))
            production_ratio_stock[production_ratio_stock > 1] = 1
            if (production_ratio_stock < 1).any():
                production_max = np.tile(production_opt, (self.n_sectors, 1)) * production_ratio_stock
                assert not (np.min(production_max,axis=0) < 0).any()
                self.production = np.min(production_max, axis=0)
            else:
                assert not (production_opt < 0).any()
                self.production = production_opt
        else:
            if self.in_shortage:
                self.in_shortage = False
                logger.info('All industries exited shortage regime. (current temporal unit:{})'.format(current_temporal_unit))
            assert not (production_opt < 0).any()
            self.production = production_opt
        return stock_constraint

    def calc_matrix_stock_gap(self, matrix_stock_goal) -> np.ndarray:
        matrix_stock_gap = super().calc_matrix_stock_gap(matrix_stock_goal)
        return np.expand_dims(self.restoration_tau, axis=1) * matrix_stock_gap



"""
This type stub file was generated by pyright.
"""

import numpy as np
from typing import Iterable, Literal

"""
Operator classes for eval.
"""
REDUCTIONS = ...
_unary_math_ops = ...
_binary_math_ops = ...
MATHOPS = ...
LOCAL_TAG = ...
class Term:
    def __new__(cls, name, env, side=..., encoding=...): # -> Constant | Self@Term:
        ...
    
    is_local: bool
    def __init__(self, name, env, side=..., encoding=...) -> None:
        ...
    
    @property
    def local_name(self) -> str:
        ...
    
    def __repr__(self) -> str:
        ...
    
    def __call__(self, *args, **kwargs):
        ...
    
    def evaluate(self, *args, **kwargs) -> Term:
        ...
    
    def update(self, value) -> None:
        """
        search order for local (i.e., @variable) variables:

        scope, key_variable
        [('locals', 'local_name'),
         ('globals', 'local_name'),
         ('locals', 'key'),
         ('globals', 'key')]
        """
        ...
    
    @property
    def is_scalar(self) -> bool:
        ...
    
    @property
    def type(self): # -> Any:
        ...
    
    return_type = ...
    @property
    def raw(self) -> str:
        ...
    
    @property
    def is_datetime(self) -> bool:
        ...
    
    @property
    def value(self):
        ...
    
    @value.setter
    def value(self, new_value): # -> None:
        ...
    
    @property
    def name(self): # -> Unknown:
        ...
    
    @property
    def ndim(self) -> int:
        ...
    


class Constant(Term):
    def __init__(self, value, env, side=..., encoding=...) -> None:
        ...
    
    @property
    def name(self):
        ...
    
    def __repr__(self) -> str:
        ...
    


_bool_op_map = ...
class Op:
    """
    Hold an operator of arbitrary arity.
    """
    op: str
    def __init__(self, op: str, operands: Iterable[Term | Op], encoding=...) -> None:
        ...
    
    def __iter__(self): # -> Iterator[Term | Op]:
        ...
    
    def __repr__(self) -> str:
        """
        Print a generic n-ary operator and its operands using infix notation.
        """
        ...
    
    @property
    def return_type(self): # -> Type[bool_] | dtype[Any]:
        ...
    
    @property
    def has_invalid_return_type(self) -> bool:
        ...
    
    @property
    def operand_types(self): # -> frozenset[Unknown]:
        ...
    
    @property
    def is_scalar(self) -> bool:
        ...
    
    @property
    def is_datetime(self) -> bool:
        ...
    


CMP_OPS_SYMS = ...
_cmp_ops_funcs = ...
_cmp_ops_dict = ...
BOOL_OPS_SYMS = ...
_bool_ops_funcs = ...
_bool_ops_dict = ...
ARITH_OPS_SYMS = ...
_arith_ops_funcs = ...
_arith_ops_dict = ...
SPECIAL_CASE_ARITH_OPS_SYMS = ...
_special_case_arith_ops_funcs = ...
_special_case_arith_ops_dict = ...
_binary_ops_dict = ...
def is_term(obj) -> bool:
    ...

class BinOp(Op):
    """
    Hold a binary operator and its operands.

    Parameters
    ----------
    op : str
    lhs : Term or Op
    rhs : Term or Op
    """
    def __init__(self, op: str, lhs, rhs) -> None:
        ...
    
    def __call__(self, env):
        """
        Recursively evaluate an expression in Python space.

        Parameters
        ----------
        env : Scope

        Returns
        -------
        object
            The result of an evaluated expression.
        """
        ...
    
    def evaluate(self, env, engine: str, parser, term_type, eval_in_python):
        """
        Evaluate a binary operation *before* being passed to the engine.

        Parameters
        ----------
        env : Scope
        engine : str
        parser : str
        term_type : type
        eval_in_python : list

        Returns
        -------
        term_type
            The "pre-evaluated" expression as an instance of ``term_type``
        """
        ...
    
    def convert_values(self) -> None:
        """
        Convert datetimes to a comparable value in an expression.
        """
        ...
    


def isnumeric(dtype) -> bool:
    ...

class Div(BinOp):
    """
    Div operator to special case casting.

    Parameters
    ----------
    lhs, rhs : Term or Op
        The Terms or Ops in the ``/`` expression.
    """
    def __init__(self, lhs, rhs) -> None:
        ...
    


UNARY_OPS_SYMS = ...
_unary_ops_funcs = ...
_unary_ops_dict = ...
class UnaryOp(Op):
    """
    Hold a unary operator and its operands.

    Parameters
    ----------
    op : str
        The token used to represent the operator.
    operand : Term or Op
        The Term or Op operand to the operator.

    Raises
    ------
    ValueError
        * If no function associated with the passed operator token is found.
    """
    def __init__(self, op: Literal["+", "-", "~", "not"], operand) -> None:
        ...
    
    def __call__(self, env) -> MathCall:
        ...
    
    def __repr__(self) -> str:
        ...
    
    @property
    def return_type(self) -> np.dtype:
        ...
    


class MathCall(Op):
    def __init__(self, func, args) -> None:
        ...
    
    def __call__(self, env):
        ...
    
    def __repr__(self) -> str:
        ...
    


class FuncNode:
    def __init__(self, name: str) -> None:
        ...
    
    def __call__(self, *args): # -> MathCall:
        ...
    





"""
This type stub file was generated by pyright.
"""

from pandas.api import extensions, indexers, interchange, types

""" public toolkit API """
__all__ = ["interchange", "extensions", "indexers", "types"]



"""
This type stub file was generated by pyright.
"""

from typing import Any, TYPE_CHECKING
from pandas._typing import FilePath, ReadBuffer, Scalar, StorageOptions, WriteExcelBuffer
from pandas.util._decorators import doc
from pandas.core.shared_docs import _shared_docs
from pandas.io.excel._base import BaseExcelReader, ExcelWriter
from openpyxl.workbook import Workbook

if TYPE_CHECKING:
    ...
class OpenpyxlWriter(ExcelWriter):
    _engine = ...
    _supported_extensions = ...
    def __init__(self, path: FilePath | WriteExcelBuffer | ExcelWriter, engine: str | None = ..., date_format: str | None = ..., datetime_format: str | None = ..., mode: str = ..., storage_options: StorageOptions = ..., if_sheet_exists: str | None = ..., engine_kwargs: dict[str, Any] | None = ..., **kwargs) -> None:
        ...
    
    @property
    def book(self) -> Workbook:
        """
        Book instance of class openpyxl.workbook.Workbook.

        This attribute can be used to access engine-specific features.
        """
        ...
    
    @property
    def sheets(self) -> dict[str, Any]:
        """Mapping of sheet names to sheet objects."""
        ...
    


class OpenpyxlReader(BaseExcelReader):
    @doc(storage_options=_shared_docs["storage_options"])
    def __init__(self, filepath_or_buffer: FilePath | ReadBuffer[bytes], storage_options: StorageOptions = ...) -> None:
        """
        Reader using openpyxl engine.

        Parameters
        ----------
        filepath_or_buffer : str, path object or Workbook
            Object to be parsed.
        {storage_options}
        """
        ...
    
    def load_workbook(self, filepath_or_buffer: FilePath | ReadBuffer[bytes]):
        ...
    
    @property
    def sheet_names(self) -> list[str]:
        ...
    
    def get_sheet_by_name(self, name: str):
        ...
    
    def get_sheet_by_index(self, index: int):
        ...
    
    def get_sheet_data(self, sheet, convert_float: bool, file_rows_needed: int | None = ...) -> list[list[Scalar]]:
        ...
    





"""
This type stub file was generated by pyright.
"""

import numpy as np
from pandas._typing import npt

def array_strptime(values: npt.NDArray[np.object_], fmt: str | None, exact: bool = ..., errors: str = ...) -> tuple[np.ndarray, np.ndarray]:
    ...




"""
This type stub file was generated by pyright.
"""

"""
Templating for ops docstrings
"""
def make_flex_doc(op_name: str, typ: str) -> str:
    """
    Make the appropriate substitutions for the given operation and class-typ
    into either _flex_doc_SERIES or _flex_doc_FRAME to return the docstring
    to attach to a generated method.

    Parameters
    ----------
    op_name : str {'__add__', '__sub__', ... '__eq__', '__ne__', ...}
    typ : str {series, 'dataframe']}

    Returns
    -------
    doc : str
    """
    ...

_common_examples_algebra_SERIES = ...
_common_examples_comparison_SERIES = ...
_add_example_SERIES = ...
_sub_example_SERIES = ...
_mul_example_SERIES = ...
_div_example_SERIES = ...
_floordiv_example_SERIES = ...
_divmod_example_SERIES = ...
_mod_example_SERIES = ...
_pow_example_SERIES = ...
_ne_example_SERIES = ...
_eq_example_SERIES = ...
_lt_example_SERIES = ...
_le_example_SERIES = ...
_gt_example_SERIES = ...
_ge_example_SERIES = ...
_returns_series = ...
_returns_tuple = ...
_op_descriptions: dict[str, dict[str, str | None]] = ...
_py_num_ref = ...
_op_names = ...
_flex_doc_SERIES = ...
_see_also_reverse_SERIES = ...
_flex_doc_FRAME = ...
_flex_comp_doc_FRAME = ...



"""
This type stub file was generated by pyright.
"""

import numpy as np
from datetime import date as _date, datetime, time as _time, timedelta, tzinfo as _tzinfo
from time import struct_time
from typing import ClassVar, TypeVar, overload
from pandas._libs.tslibs import BaseOffset, NaTType, Period, Tick, Timedelta

_DatetimeT = TypeVar("_DatetimeT", bound=datetime)
def integer_op_not_supported(obj: object) -> TypeError:
    ...

class Timestamp(datetime):
    min: ClassVar[Timestamp]
    max: ClassVar[Timestamp]
    resolution: ClassVar[Timedelta]
    value: int
    def __new__(cls: type[_DatetimeT], ts_input: np.integer | float | str | _date | datetime | np.datetime64 = ..., freq: int | None | str | BaseOffset = ..., tz: str | _tzinfo | None | int = ..., unit: str | int | None = ..., year: int | None = ..., month: int | None = ..., day: int | None = ..., hour: int | None = ..., minute: int | None = ..., second: int | None = ..., microsecond: int | None = ..., nanosecond: int | None = ..., tzinfo: _tzinfo | None = ..., *, fold: int | None = ...) -> _DatetimeT | NaTType:
        ...
    
    @property
    def year(self) -> int:
        ...
    
    @property
    def month(self) -> int:
        ...
    
    @property
    def day(self) -> int:
        ...
    
    @property
    def hour(self) -> int:
        ...
    
    @property
    def minute(self) -> int:
        ...
    
    @property
    def second(self) -> int:
        ...
    
    @property
    def microsecond(self) -> int:
        ...
    
    @property
    def tzinfo(self) -> _tzinfo | None:
        ...
    
    @property
    def tz(self) -> _tzinfo | None:
        ...
    
    @property
    def fold(self) -> int:
        ...
    
    @classmethod
    def fromtimestamp(cls: type[_DatetimeT], ts: float, tz: _tzinfo | None = ...) -> _DatetimeT:
        ...
    
    @classmethod
    def utcfromtimestamp(cls: type[_DatetimeT], ts: float) -> _DatetimeT:
        ...
    
    @classmethod
    def today(cls: type[_DatetimeT], tz: _tzinfo | str | None = ...) -> _DatetimeT:
        ...
    
    @classmethod
    def fromordinal(cls: type[_DatetimeT], ordinal: int, freq: str | BaseOffset | None = ..., tz: _tzinfo | str | None = ...) -> _DatetimeT:
        ...
    
    @classmethod
    def now(cls: type[_DatetimeT], tz: _tzinfo | str | None = ...) -> _DatetimeT:
        ...
    
    @classmethod
    def utcnow(cls: type[_DatetimeT]) -> _DatetimeT:
        ...
    
    @classmethod
    def combine(cls, date: _date, time: _time) -> datetime:
        ...
    
    @classmethod
    def fromisoformat(cls: type[_DatetimeT], date_string: str) -> _DatetimeT:
        ...
    
    def strftime(self, format: str) -> str:
        ...
    
    def __format__(self, fmt: str) -> str:
        ...
    
    def toordinal(self) -> int:
        ...
    
    def timetuple(self) -> struct_time:
        ...
    
    def timestamp(self) -> float:
        ...
    
    def utctimetuple(self) -> struct_time:
        ...
    
    def date(self) -> _date:
        ...
    
    def time(self) -> _time:
        ...
    
    def timetz(self) -> _time:
        ...
    
    def replace(self: _DatetimeT, year: int | None = ..., month: int | None = ..., day: int | None = ..., hour: int | None = ..., minute: int | None = ..., second: int | None = ..., microsecond: int | None = ..., nanosecond: int | None = ..., tzinfo: _tzinfo | type[object] | None = ..., fold: int | None = ...) -> _DatetimeT:
        ...
    
    def astimezone(self: _DatetimeT, tz: _tzinfo | None) -> _DatetimeT:
        ...
    
    def ctime(self) -> str:
        ...
    
    def isoformat(self, sep: str = ..., timespec: str = ...) -> str:
        ...
    
    @classmethod
    def strptime(cls, date_string: str, format: str) -> datetime:
        ...
    
    def utcoffset(self) -> timedelta | None:
        ...
    
    def tzname(self) -> str | None:
        ...
    
    def dst(self) -> timedelta | None:
        ...
    
    def __le__(self, other: datetime) -> bool:
        ...
    
    def __lt__(self, other: datetime) -> bool:
        ...
    
    def __ge__(self, other: datetime) -> bool:
        ...
    
    def __gt__(self, other: datetime) -> bool:
        ...
    
    @overload
    def __add__(self, other: np.ndarray) -> np.ndarray:
        ...
    
    @overload
    def __add__(self: _DatetimeT, other: timedelta | np.timedelta64 | Tick) -> _DatetimeT:
        ...
    
    def __radd__(self: _DatetimeT, other: timedelta) -> _DatetimeT:
        ...
    
    @overload
    def __sub__(self, other: datetime) -> Timedelta:
        ...
    
    @overload
    def __sub__(self: _DatetimeT, other: timedelta | np.timedelta64 | Tick) -> _DatetimeT:
        ...
    
    def __hash__(self) -> int:
        ...
    
    def weekday(self) -> int:
        ...
    
    def isoweekday(self) -> int:
        ...
    
    def isocalendar(self) -> tuple[int, int, int]:
        ...
    
    @property
    def is_leap_year(self) -> bool:
        ...
    
    @property
    def is_month_start(self) -> bool:
        ...
    
    @property
    def is_quarter_start(self) -> bool:
        ...
    
    @property
    def is_year_start(self) -> bool:
        ...
    
    @property
    def is_month_end(self) -> bool:
        ...
    
    @property
    def is_quarter_end(self) -> bool:
        ...
    
    @property
    def is_year_end(self) -> bool:
        ...
    
    def to_pydatetime(self, warn: bool = ...) -> datetime:
        ...
    
    def to_datetime64(self) -> np.datetime64:
        ...
    
    def to_period(self, freq: BaseOffset | str | None = ...) -> Period:
        ...
    
    def to_julian_date(self) -> np.float64:
        ...
    
    @property
    def asm8(self) -> np.datetime64:
        ...
    
    def tz_convert(self: _DatetimeT, tz: _tzinfo | str | None) -> _DatetimeT:
        ...
    
    def tz_localize(self: _DatetimeT, tz: _tzinfo | str | None, ambiguous: str = ..., nonexistent: str = ...) -> _DatetimeT:
        ...
    
    def normalize(self: _DatetimeT) -> _DatetimeT:
        ...
    
    def round(self: _DatetimeT, freq: str, ambiguous: bool | str = ..., nonexistent: str = ...) -> _DatetimeT:
        ...
    
    def floor(self: _DatetimeT, freq: str, ambiguous: bool | str = ..., nonexistent: str = ...) -> _DatetimeT:
        ...
    
    def ceil(self: _DatetimeT, freq: str, ambiguous: bool | str = ..., nonexistent: str = ...) -> _DatetimeT:
        ...
    
    def day_name(self, locale: str | None = ...) -> str:
        ...
    
    def month_name(self, locale: str | None = ...) -> str:
        ...
    
    @property
    def day_of_week(self) -> int:
        ...
    
    @property
    def dayofweek(self) -> int:
        ...
    
    @property
    def day_of_year(self) -> int:
        ...
    
    @property
    def dayofyear(self) -> int:
        ...
    
    @property
    def quarter(self) -> int:
        ...
    
    @property
    def week(self) -> int:
        ...
    
    def to_numpy(self, dtype: np.dtype | None = ..., copy: bool = ...) -> np.datetime64:
        ...
    
    @property
    def days_in_month(self) -> int:
        ...
    
    @property
    def daysinmonth(self) -> int:
        ...
    





"""
This type stub file was generated by pyright.
"""

""" io on the clipboard """
def read_clipboard(sep: str = ..., **kwargs): # -> TextFileReader:
    r"""
    Read text from clipboard and pass to read_csv.

    Parameters
    ----------
    sep : str, default '\s+'
        A string or regex delimiter. The default of '\s+' denotes
        one or more whitespace characters.

    **kwargs
        See read_csv for the full argument list.

    Returns
    -------
    DataFrame
        A parsed DataFrame object.
    """
    ...

def to_clipboard(obj, excel: bool | None = ..., sep: str | None = ..., **kwargs) -> None:
    """
    Attempt to write text representation of object to the system clipboard
    The clipboard can be then pasted into Excel for example.

    Parameters
    ----------
    obj : the object to write to the clipboard
    excel : bool, defaults to True
            if True, use the provided separator, writing in a csv
            format for allowing easy pasting into excel.
            if False, write a string representation of the object
            to the clipboard
    sep : optional, defaults to tab
    other keywords are passed to to_csv

    Notes
    -----
    Requirements for your platform
      - Linux: xclip, or xsel (with PyQt4 modules)
      - Windows:
      - OS X:
    """
    ...




"""
This type stub file was generated by pyright.
"""

from typing import Any
from pandas._typing import ArrayLike

"""
Functions for defining unary operations.
"""
def should_extension_dispatch(left: ArrayLike, right: Any) -> bool:
    """
    Identify cases where Series operation should dispatch to ExtensionArray method.

    Parameters
    ----------
    left : np.ndarray or ExtensionArray
    right : object

    Returns
    -------
    bool
    """
    ...




"""
This type stub file was generated by pyright.
"""

from typing import Sequence, overload
from pandas._typing import AnyArrayLike, DataFrame, Index, Series

cache_readonly = property
class AxisProperty:
    axis: int
    def __init__(self, axis: int = ..., doc: str = ...) -> None:
        ...
    
    @overload
    def __get__(self, obj: DataFrame | Series, type) -> Index:
        ...
    
    @overload
    def __get__(self, obj: None, type) -> AxisProperty:
        ...
    
    def __set__(self, obj: DataFrame | Series, value: AnyArrayLike | Sequence) -> None:
        ...
    





"""
This type stub file was generated by pyright.
"""

import numpy as np
from typing import Any, Literal, TYPE_CHECKING, TypeVar, overload
from pandas._libs.arrays import NDArrayBacked
from pandas._typing import ArrayLike, Dtype, F, NumpySorter, NumpyValueArrayLike, PositionalIndexer2D, PositionalIndexerTuple, ScalarIndexer, SequenceIndexer, TakeIndexer, npt
from pandas.util._decorators import doc
from pandas.core.array_algos.transforms import shift
from pandas.core.arrays.base import ExtensionArray
from pandas import Series

NDArrayBackedExtensionArrayT = TypeVar("NDArrayBackedExtensionArrayT", bound="NDArrayBackedExtensionArray")
if TYPE_CHECKING:
    ...
def ravel_compat(meth: F) -> F:
    """
    Decorator to ravel a 2D array before passing it to a cython operation,
    then reshape the result to our own shape.
    """
    ...

class NDArrayBackedExtensionArray(NDArrayBacked, ExtensionArray):
    """
    ExtensionArray that is backed by a single NumPy ndarray.
    """
    _ndarray: np.ndarray
    _internal_fill_value: Any
    def view(self, dtype: Dtype | None = ...) -> ArrayLike:
        ...
    
    def take(self: NDArrayBackedExtensionArrayT, indices: TakeIndexer, *, allow_fill: bool = ..., fill_value: Any = ..., axis: int = ...) -> NDArrayBackedExtensionArrayT:
        ...
    
    def equals(self, other) -> bool:
        ...
    
    def argmin(self, axis: int = ..., skipna: bool = ...): # -> NDArray[Any] | Any | int:
        ...
    
    def argmax(self, axis: int = ..., skipna: bool = ...): # -> NDArray[Any] | Any | int:
        ...
    
    def unique(self: NDArrayBackedExtensionArrayT) -> NDArrayBackedExtensionArrayT:
        ...
    
    @doc(ExtensionArray.searchsorted)
    def searchsorted(self, value: NumpyValueArrayLike | ExtensionArray, side: Literal["left", "right"] = ..., sorter: NumpySorter = ...) -> npt.NDArray[np.intp] | np.intp:
        ...
    
    @doc(ExtensionArray.shift)
    def shift(self, periods=..., fill_value=..., axis=...):
        ...
    
    def __setitem__(self, key, value) -> None:
        ...
    
    @overload
    def __getitem__(self, key: ScalarIndexer) -> Any:
        ...
    
    @overload
    def __getitem__(self: NDArrayBackedExtensionArrayT, key: SequenceIndexer | PositionalIndexerTuple) -> NDArrayBackedExtensionArrayT:
        ...
    
    def __getitem__(self: NDArrayBackedExtensionArrayT, key: PositionalIndexer2D) -> NDArrayBackedExtensionArrayT | Any:
        ...
    
    @doc(ExtensionArray.fillna)
    def fillna(self: NDArrayBackedExtensionArrayT, value=..., method=..., limit=...) -> NDArrayBackedExtensionArrayT:
        ...
    
    def insert(self: NDArrayBackedExtensionArrayT, loc: int, item) -> NDArrayBackedExtensionArrayT:
        """
        Make new ExtensionArray inserting new item at location. Follows
        Python list.append semantics for negative values.

        Parameters
        ----------
        loc : int
        item : object

        Returns
        -------
        type(self)
        """
        ...
    
    def value_counts(self, dropna: bool = ...) -> Series:
        """
        Return a Series containing counts of unique values.

        Parameters
        ----------
        dropna : bool, default True
            Don't include counts of NA values.

        Returns
        -------
        Series
        """
        ...
    





"""
This type stub file was generated by pyright.
"""

__all__ = ["concat", "crosstab", "cut", "from_dummies", "get_dummies", "lreshape", "melt", "merge", "merge_asof", "merge_ordered", "pivot", "pivot_table", "qcut", "wide_to_long"]



"""
This type stub file was generated by pyright.
"""

import copy
import numpy as np
from typing import Hashable, Sequence, TYPE_CHECKING
from pandas._typing import AnyArrayLike, IndexLabel, Suffixes, npt
from pandas.util._decorators import Appender, Substitution
from pandas import DataFrame, Index, MultiIndex, Series
from pandas.core.frame import _merge_doc

"""
SQL-style merge routines
"""
if TYPE_CHECKING:
    ...
@Substitution("\nleft : DataFrame or named Series")
@Appender(_merge_doc, indents=0)
def merge(left: DataFrame | Series, right: DataFrame | Series, how: str = ..., on: IndexLabel | None = ..., left_on: IndexLabel | None = ..., right_on: IndexLabel | None = ..., left_index: bool = ..., right_index: bool = ..., sort: bool = ..., suffixes: Suffixes = ..., copy: bool = ..., indicator: bool = ..., validate: str | None = ...) -> DataFrame:
    ...

if __debug__:
    ...
def merge_ordered(left: DataFrame, right: DataFrame, on: IndexLabel | None = ..., left_on: IndexLabel | None = ..., right_on: IndexLabel | None = ..., left_by=..., right_by=..., fill_method: str | None = ..., suffixes: Suffixes = ..., how: str = ...) -> DataFrame:
    """
    Perform a merge for ordered data with optional filling/interpolation.

    Designed for ordered data like time series data. Optionally
    perform group-wise merge (see examples).

    Parameters
    ----------
    left : DataFrame
    right : DataFrame
    on : label or list
        Field names to join on. Must be found in both DataFrames.
    left_on : label or list, or array-like
        Field names to join on in left DataFrame. Can be a vector or list of
        vectors of the length of the DataFrame to use a particular vector as
        the join key instead of columns.
    right_on : label or list, or array-like
        Field names to join on in right DataFrame or vector/list of vectors per
        left_on docs.
    left_by : column name or list of column names
        Group left DataFrame by group columns and merge piece by piece with
        right DataFrame.
    right_by : column name or list of column names
        Group right DataFrame by group columns and merge piece by piece with
        left DataFrame.
    fill_method : {'ffill', None}, default None
        Interpolation method for data.
    suffixes : list-like, default is ("_x", "_y")
        A length-2 sequence where each element is optionally a string
        indicating the suffix to add to overlapping column names in
        `left` and `right` respectively. Pass a value of `None` instead
        of a string to indicate that the column name from `left` or
        `right` should be left as-is, with no suffix. At least one of the
        values must not be None.

        .. versionchanged:: 0.25.0
    how : {'left', 'right', 'outer', 'inner'}, default 'outer'
        * left: use only keys from left frame (SQL: left outer join)
        * right: use only keys from right frame (SQL: right outer join)
        * outer: use union of keys from both frames (SQL: full outer join)
        * inner: use intersection of keys from both frames (SQL: inner join).

    Returns
    -------
    DataFrame
        The merged DataFrame output type will the be same as
        'left', if it is a subclass of DataFrame.

    See Also
    --------
    merge : Merge with a database-style join.
    merge_asof : Merge on nearest keys.

    Examples
    --------
    >>> df1 = pd.DataFrame(
    ...     {
    ...         "key": ["a", "c", "e", "a", "c", "e"],
    ...         "lvalue": [1, 2, 3, 1, 2, 3],
    ...         "group": ["a", "a", "a", "b", "b", "b"]
    ...     }
    ... )
    >>> df1
          key  lvalue group
    0   a       1     a
    1   c       2     a
    2   e       3     a
    3   a       1     b
    4   c       2     b
    5   e       3     b

    >>> df2 = pd.DataFrame({"key": ["b", "c", "d"], "rvalue": [1, 2, 3]})
    >>> df2
          key  rvalue
    0   b       1
    1   c       2
    2   d       3

    >>> merge_ordered(df1, df2, fill_method="ffill", left_by="group")
      key  lvalue group  rvalue
    0   a       1     a     NaN
    1   b       1     a     1.0
    2   c       2     a     2.0
    3   d       2     a     3.0
    4   e       3     a     3.0
    5   a       1     b     NaN
    6   b       1     b     1.0
    7   c       2     b     2.0
    8   d       2     b     3.0
    9   e       3     b     3.0
    """
    ...

def merge_asof(left: DataFrame | Series, right: DataFrame | Series, on: IndexLabel | None = ..., left_on: IndexLabel | None = ..., right_on: IndexLabel | None = ..., left_index: bool = ..., right_index: bool = ..., by=..., left_by=..., right_by=..., suffixes: Suffixes = ..., tolerance=..., allow_exact_matches: bool = ..., direction: str = ...) -> DataFrame:
    """
    Perform a merge by key distance.

    This is similar to a left-join except that we match on nearest
    key rather than equal keys. Both DataFrames must be sorted by the key.

    For each row in the left DataFrame:

      - A "backward" search selects the last row in the right DataFrame whose
        'on' key is less than or equal to the left's key.

      - A "forward" search selects the first row in the right DataFrame whose
        'on' key is greater than or equal to the left's key.

      - A "nearest" search selects the row in the right DataFrame whose 'on'
        key is closest in absolute distance to the left's key.

    The default is "backward" and is compatible in versions below 0.20.0.
    The direction parameter was added in version 0.20.0 and introduces
    "forward" and "nearest".

    Optionally match on equivalent keys with 'by' before searching with 'on'.

    Parameters
    ----------
    left : DataFrame or named Series
    right : DataFrame or named Series
    on : label
        Field name to join on. Must be found in both DataFrames.
        The data MUST be ordered. Furthermore this must be a numeric column,
        such as datetimelike, integer, or float. On or left_on/right_on
        must be given.
    left_on : label
        Field name to join on in left DataFrame.
    right_on : label
        Field name to join on in right DataFrame.
    left_index : bool
        Use the index of the left DataFrame as the join key.
    right_index : bool
        Use the index of the right DataFrame as the join key.
    by : column name or list of column names
        Match on these columns before performing merge operation.
    left_by : column name
        Field names to match on in the left DataFrame.
    right_by : column name
        Field names to match on in the right DataFrame.
    suffixes : 2-length sequence (tuple, list, ...)
        Suffix to apply to overlapping column names in the left and right
        side, respectively.
    tolerance : int or Timedelta, optional, default None
        Select asof tolerance within this range; must be compatible
        with the merge index.
    allow_exact_matches : bool, default True

        - If True, allow matching with the same 'on' value
          (i.e. less-than-or-equal-to / greater-than-or-equal-to)
        - If False, don't match the same 'on' value
          (i.e., strictly less-than / strictly greater-than).

    direction : 'backward' (default), 'forward', or 'nearest'
        Whether to search for prior, subsequent, or closest matches.

    Returns
    -------
    merged : DataFrame

    See Also
    --------
    merge : Merge with a database-style join.
    merge_ordered : Merge with optional filling/interpolation.

    Examples
    --------
    >>> left = pd.DataFrame({"a": [1, 5, 10], "left_val": ["a", "b", "c"]})
    >>> left
        a left_val
    0   1        a
    1   5        b
    2  10        c

    >>> right = pd.DataFrame({"a": [1, 2, 3, 6, 7], "right_val": [1, 2, 3, 6, 7]})
    >>> right
       a  right_val
    0  1          1
    1  2          2
    2  3          3
    3  6          6
    4  7          7

    >>> pd.merge_asof(left, right, on="a")
        a left_val  right_val
    0   1        a          1
    1   5        b          3
    2  10        c          7

    >>> pd.merge_asof(left, right, on="a", allow_exact_matches=False)
        a left_val  right_val
    0   1        a        NaN
    1   5        b        3.0
    2  10        c        7.0

    >>> pd.merge_asof(left, right, on="a", direction="forward")
        a left_val  right_val
    0   1        a        1.0
    1   5        b        6.0
    2  10        c        NaN

    >>> pd.merge_asof(left, right, on="a", direction="nearest")
        a left_val  right_val
    0   1        a          1
    1   5        b          6
    2  10        c          7

    We can use indexed DataFrames as well.

    >>> left = pd.DataFrame({"left_val": ["a", "b", "c"]}, index=[1, 5, 10])
    >>> left
       left_val
    1         a
    5         b
    10        c

    >>> right = pd.DataFrame({"right_val": [1, 2, 3, 6, 7]}, index=[1, 2, 3, 6, 7])
    >>> right
       right_val
    1          1
    2          2
    3          3
    6          6
    7          7

    >>> pd.merge_asof(left, right, left_index=True, right_index=True)
       left_val  right_val
    1         a          1
    5         b          3
    10        c          7

    Here is a real-world times-series example

    >>> quotes = pd.DataFrame(
    ...     {
    ...         "time": [
    ...             pd.Timestamp("2016-05-25 13:30:00.023"),
    ...             pd.Timestamp("2016-05-25 13:30:00.023"),
    ...             pd.Timestamp("2016-05-25 13:30:00.030"),
    ...             pd.Timestamp("2016-05-25 13:30:00.041"),
    ...             pd.Timestamp("2016-05-25 13:30:00.048"),
    ...             pd.Timestamp("2016-05-25 13:30:00.049"),
    ...             pd.Timestamp("2016-05-25 13:30:00.072"),
    ...             pd.Timestamp("2016-05-25 13:30:00.075")
    ...         ],
    ...         "ticker": [
    ...                "GOOG",
    ...                "MSFT",
    ...                "MSFT",
    ...                "MSFT",
    ...                "GOOG",
    ...                "AAPL",
    ...                "GOOG",
    ...                "MSFT"
    ...            ],
    ...            "bid": [720.50, 51.95, 51.97, 51.99, 720.50, 97.99, 720.50, 52.01],
    ...            "ask": [720.93, 51.96, 51.98, 52.00, 720.93, 98.01, 720.88, 52.03]
    ...     }
    ... )
    >>> quotes
                         time ticker     bid     ask
    0 2016-05-25 13:30:00.023   GOOG  720.50  720.93
    1 2016-05-25 13:30:00.023   MSFT   51.95   51.96
    2 2016-05-25 13:30:00.030   MSFT   51.97   51.98
    3 2016-05-25 13:30:00.041   MSFT   51.99   52.00
    4 2016-05-25 13:30:00.048   GOOG  720.50  720.93
    5 2016-05-25 13:30:00.049   AAPL   97.99   98.01
    6 2016-05-25 13:30:00.072   GOOG  720.50  720.88
    7 2016-05-25 13:30:00.075   MSFT   52.01   52.03

    >>> trades = pd.DataFrame(
    ...        {
    ...            "time": [
    ...                pd.Timestamp("2016-05-25 13:30:00.023"),
    ...                pd.Timestamp("2016-05-25 13:30:00.038"),
    ...                pd.Timestamp("2016-05-25 13:30:00.048"),
    ...                pd.Timestamp("2016-05-25 13:30:00.048"),
    ...                pd.Timestamp("2016-05-25 13:30:00.048")
    ...            ],
    ...            "ticker": ["MSFT", "MSFT", "GOOG", "GOOG", "AAPL"],
    ...            "price": [51.95, 51.95, 720.77, 720.92, 98.0],
    ...            "quantity": [75, 155, 100, 100, 100]
    ...        }
    ...    )
    >>> trades
                         time ticker   price  quantity
    0 2016-05-25 13:30:00.023   MSFT   51.95        75
    1 2016-05-25 13:30:00.038   MSFT   51.95       155
    2 2016-05-25 13:30:00.048   GOOG  720.77       100
    3 2016-05-25 13:30:00.048   GOOG  720.92       100
    4 2016-05-25 13:30:00.048   AAPL   98.00       100

    By default we are taking the asof of the quotes

    >>> pd.merge_asof(trades, quotes, on="time", by="ticker")
                         time ticker   price  quantity     bid     ask
    0 2016-05-25 13:30:00.023   MSFT   51.95        75   51.95   51.96
    1 2016-05-25 13:30:00.038   MSFT   51.95       155   51.97   51.98
    2 2016-05-25 13:30:00.048   GOOG  720.77       100  720.50  720.93
    3 2016-05-25 13:30:00.048   GOOG  720.92       100  720.50  720.93
    4 2016-05-25 13:30:00.048   AAPL   98.00       100     NaN     NaN

    We only asof within 2ms between the quote time and the trade time

    >>> pd.merge_asof(
    ...     trades, quotes, on="time", by="ticker", tolerance=pd.Timedelta("2ms")
    ... )
                         time ticker   price  quantity     bid     ask
    0 2016-05-25 13:30:00.023   MSFT   51.95        75   51.95   51.96
    1 2016-05-25 13:30:00.038   MSFT   51.95       155     NaN     NaN
    2 2016-05-25 13:30:00.048   GOOG  720.77       100  720.50  720.93
    3 2016-05-25 13:30:00.048   GOOG  720.92       100  720.50  720.93
    4 2016-05-25 13:30:00.048   AAPL   98.00       100     NaN     NaN

    We only asof within 10ms between the quote time and the trade time
    and we exclude exact matches on time. However *prior* data will
    propagate forward

    >>> pd.merge_asof(
    ...     trades,
    ...     quotes,
    ...     on="time",
    ...     by="ticker",
    ...     tolerance=pd.Timedelta("10ms"),
    ...     allow_exact_matches=False
    ... )
                         time ticker   price  quantity     bid     ask
    0 2016-05-25 13:30:00.023   MSFT   51.95        75     NaN     NaN
    1 2016-05-25 13:30:00.038   MSFT   51.95       155   51.97   51.98
    2 2016-05-25 13:30:00.048   GOOG  720.77       100     NaN     NaN
    3 2016-05-25 13:30:00.048   GOOG  720.92       100     NaN     NaN
    4 2016-05-25 13:30:00.048   AAPL   98.00       100     NaN     NaN
    """
    ...

class _MergeOperation:
    """
    Perform a database (SQL) merge operation between two DataFrame or Series
    objects using either columns as keys or their row indexes
    """
    _merge_type = ...
    how: str
    on: IndexLabel | None
    left_on: Sequence[Hashable | AnyArrayLike]
    right_on: Sequence[Hashable | AnyArrayLike]
    left_index: bool
    right_index: bool
    axis: int
    bm_axis: int
    sort: bool
    suffixes: Suffixes
    copy: bool
    indicator: bool
    validate: str | None
    def __init__(self, left: DataFrame | Series, right: DataFrame | Series, how: str = ..., on: IndexLabel | None = ..., left_on: IndexLabel | None = ..., right_on: IndexLabel | None = ..., axis: int = ..., left_index: bool = ..., right_index: bool = ..., sort: bool = ..., suffixes: Suffixes = ..., indicator: bool = ..., validate: str | None = ...) -> None:
        ...
    
    def get_result(self, copy: bool = ...) -> DataFrame:
        ...
    


def get_join_indexers(left_keys, right_keys, sort: bool = ..., how: str = ..., **kwargs) -> tuple[npt.NDArray[np.intp], npt.NDArray[np.intp]]:
    """

    Parameters
    ----------
    left_keys : ndarray, Index, Series
    right_keys : ndarray, Index, Series
    sort : bool, default False
    how : {'inner', 'outer', 'left', 'right'}, default 'inner'

    Returns
    -------
    np.ndarray[np.intp]
        Indexer into the left_keys.
    np.ndarray[np.intp]
        Indexer into the right_keys.
    """
    ...

def restore_dropped_levels_multijoin(left: MultiIndex, right: MultiIndex, dropped_level_names, join_index: Index, lindexer: npt.NDArray[np.intp], rindexer: npt.NDArray[np.intp]) -> tuple[list[Index], npt.NDArray[np.intp], list[Hashable]]:
    """
    *this is an internal non-public method*

    Returns the levels, labels and names of a multi-index to multi-index join.
    Depending on the type of join, this method restores the appropriate
    dropped levels of the joined multi-index.
    The method relies on lidx, rindexer which hold the index positions of
    left and right, where a join was feasible

    Parameters
    ----------
    left : MultiIndex
        left index
    right : MultiIndex
        right index
    dropped_level_names : str array
        list of non-common level names
    join_index : Index
        the index of the join between the
        common levels of left and right
    lindexer : np.ndarray[np.intp]
        left indexer
    rindexer : np.ndarray[np.intp]
        right indexer

    Returns
    -------
    levels : list of Index
        levels of combined multiindexes
    labels : np.ndarray[np.intp]
        labels of combined multiindexes
    names : List[Hashable]
        names of combined multiindex levels

    """
    ...

class _OrderedMerge(_MergeOperation):
    _merge_type = ...
    def __init__(self, left: DataFrame | Series, right: DataFrame | Series, on: IndexLabel | None = ..., left_on: IndexLabel | None = ..., right_on: IndexLabel | None = ..., left_index: bool = ..., right_index: bool = ..., axis: int = ..., suffixes: Suffixes = ..., fill_method: str | None = ..., how: str = ...) -> None:
        ...
    
    def get_result(self, copy: bool = ...) -> DataFrame:
        ...
    


_type_casters = ...
class _AsOfMerge(_OrderedMerge):
    _merge_type = ...
    def __init__(self, left: DataFrame | Series, right: DataFrame | Series, on: IndexLabel | None = ..., left_on: IndexLabel | None = ..., right_on: IndexLabel | None = ..., left_index: bool = ..., right_index: bool = ..., by=..., left_by=..., right_by=..., axis: int = ..., suffixes: Suffixes = ..., copy: bool = ..., fill_method: str | None = ..., how: str = ..., tolerance=..., allow_exact_matches: bool = ..., direction: str = ...) -> None:
        ...
    





"""
This type stub file was generated by pyright.
"""

import numpy as np
from typing import Callable
from pandas._libs import index as libindex
from pandas._typing import Dtype, npt
from pandas.util._decorators import cache_readonly
from pandas.core.indexes.base import Index

class NumericIndex(Index):
    """
    Immutable numeric sequence used for indexing and alignment.

    The basic object storing axis labels for all pandas objects.
    NumericIndex is a special case of `Index` with purely numpy int/uint/float labels.

    .. versionadded:: 1.4.0

    Parameters
    ----------
    data : array-like (1-dimensional)
    dtype : NumPy dtype (default: None)
    copy : bool
        Make a copy of input ndarray.
    name : object
        Name to be stored in the index.

    Attributes
    ----------
    None

    Methods
    -------
    None

    See Also
    --------
    Index : The base pandas Index type.
    Int64Index : Index of purely int64 labels (deprecated).
    UInt64Index : Index of purely uint64 labels (deprecated).
    Float64Index : Index of  purely float64 labels (deprecated).

    Notes
    -----
    An NumericIndex instance can **only** contain numpy int64/32/16/8, uint64/32/16/8 or
    float64/32/16 dtype. In particular, ``NumericIndex`` *can not* hold Pandas numeric
    dtypes (:class:`Int64Dtype`, :class:`Int32Dtype` etc.).
    """
    _typ = ...
    _values: np.ndarray
    _default_dtype: np.dtype | None = ...
    _dtype_validation_metadata: tuple[Callable[..., bool], str] = ...
    _is_numeric_dtype = ...
    _can_hold_strings = ...
    _is_backward_compat_public_numeric_index: bool = ...
    _engine_types: dict[np.dtype, type[libindex.IndexEngine]] = ...
    @cache_readonly
    def inferred_type(self) -> str:
        ...
    
    def __new__(cls, data=..., dtype: Dtype | None = ..., copy=..., name=...) -> NumericIndex:
        ...
    


_num_index_shared_docs = ...
class IntegerIndex(NumericIndex):
    """
    This is an abstract class for Int64Index, UInt64Index.
    """
    _is_backward_compat_public_numeric_index: bool = ...
    @property
    def asi8(self) -> npt.NDArray[np.int64]:
        ...
    


class Int64Index(IntegerIndex):
    _index_descr_args = ...
    __doc__ = ...
    _typ = ...
    _default_dtype = ...
    _dtype_validation_metadata = ...


class UInt64Index(IntegerIndex):
    _index_descr_args = ...
    __doc__ = ...
    _typ = ...
    _default_dtype = ...
    _dtype_validation_metadata = ...


class Float64Index(NumericIndex):
    _index_descr_args = ...
    __doc__ = ...
    _typ = ...
    _default_dtype = ...
    _dtype_validation_metadata = ...
    _is_backward_compat_public_numeric_index: bool = ...





"""
This type stub file was generated by pyright.
"""

from typing import Any, Final, Iterable
from pandas.io.formats.format import DataFrameFormatter

"""
Module for formatting output data in HTML.
"""
class HTMLFormatter:
    """
    Internal class for formatting output data in html.
    This class is intended for shared functionality between
    DataFrame.to_html() and DataFrame._repr_html_().
    Any logic in common with other output formatting methods
    should ideally be inherited from classes in format.py
    and this class responsible for only producing html markup.
    """
    indent_delta: Final = ...
    def __init__(self, formatter: DataFrameFormatter, classes: str | list[str] | tuple[str, ...] | None = ..., border: int | bool | None = ..., table_id: str | None = ..., render_links: bool = ...) -> None:
        ...
    
    def to_string(self) -> str:
        ...
    
    def render(self) -> list[str]:
        ...
    
    @property
    def should_show_dimensions(self) -> bool:
        ...
    
    @property
    def show_row_idx_names(self) -> bool:
        ...
    
    @property
    def show_col_idx_names(self) -> bool:
        ...
    
    @property
    def row_levels(self) -> int:
        ...
    
    @property
    def is_truncated(self) -> bool:
        ...
    
    @property
    def ncols(self) -> int:
        ...
    
    def write(self, s: Any, indent: int = ...) -> None:
        ...
    
    def write_th(self, s: Any, header: bool = ..., indent: int = ..., tags: str | None = ...) -> None:
        """
        Method for writing a formatted <th> cell.

        If col_space is set on the formatter then that is used for
        the value of min-width.

        Parameters
        ----------
        s : object
            The data to be written inside the cell.
        header : bool, default False
            Set to True if the <th> is for use inside <thead>.  This will
            cause min-width to be set if there is one.
        indent : int, default 0
            The indentation level of the cell.
        tags : str, default None
            Tags to include in the cell.

        Returns
        -------
        A written <th> cell.
        """
        ...
    
    def write_td(self, s: Any, indent: int = ..., tags: str | None = ...) -> None:
        ...
    
    def write_tr(self, line: Iterable, indent: int = ..., indent_delta: int = ..., header: bool = ..., align: str | None = ..., tags: dict[int, str] | None = ..., nindex_levels: int = ...) -> None:
        ...
    


class NotebookFormatter(HTMLFormatter):
    """
    Internal class for formatting output data in html for display in Jupyter
    Notebooks. This class is intended for functionality specific to
    DataFrame._repr_html_() and DataFrame.to_html(notebook=True)
    """
    def write_style(self) -> None:
        ...
    
    def render(self) -> list[str]:
        ...
    





"""
This type stub file was generated by pyright.
"""

import numpy as np
from datetime import tzinfo
from pandas._typing import npt

def format_array_from_datetime(values: npt.NDArray[np.int64], tz: tzinfo | None = ..., format: str | None = ..., na_rep: object = ..., reso: int = ...) -> npt.NDArray[np.object_]:
    ...

def array_with_unit_to_datetime(values: np.ndarray, unit: str, errors: str = ...) -> tuple[np.ndarray, tzinfo | None]:
    ...

def array_to_datetime(values: npt.NDArray[np.object_], errors: str = ..., dayfirst: bool = ..., yearfirst: bool = ..., utc: bool = ..., require_iso8601: bool = ..., allow_mixed: bool = ...) -> tuple[np.ndarray, tzinfo | None]:
    ...




"""
This type stub file was generated by pyright.
"""

from typing import TYPE_CHECKING
from pandas._typing import AggFuncType, IndexLabel
from pandas.util._decorators import Appender, Substitution, deprecate_nonkeyword_arguments
from pandas.core.frame import _shared_docs
from pandas import DataFrame

if TYPE_CHECKING:
    ...
@Substitution("\ndata : DataFrame")
@Appender(_shared_docs["pivot_table"], indents=1)
def pivot_table(data: DataFrame, values=..., index=..., columns=..., aggfunc: AggFuncType = ..., fill_value=..., margins: bool = ..., dropna: bool = ..., margins_name: str = ..., observed: bool = ..., sort: bool = ...) -> DataFrame:
    ...

@Substitution("\ndata : DataFrame")
@Appender(_shared_docs["pivot"], indents=1)
@deprecate_nonkeyword_arguments(version=None, allowed_args=["data"])
def pivot(data: DataFrame, index: IndexLabel | None = ..., columns: IndexLabel | None = ..., values: IndexLabel | None = ...) -> DataFrame:
    ...

def crosstab(index, columns, values=..., rownames=..., colnames=..., aggfunc=..., margins: bool = ..., margins_name: str = ..., dropna: bool = ..., normalize=...) -> DataFrame:
    """
    Compute a simple cross tabulation of two (or more) factors.

    By default, computes a frequency table of the factors unless an
    array of values and an aggregation function are passed.

    Parameters
    ----------
    index : array-like, Series, or list of arrays/Series
        Values to group by in the rows.
    columns : array-like, Series, or list of arrays/Series
        Values to group by in the columns.
    values : array-like, optional
        Array of values to aggregate according to the factors.
        Requires `aggfunc` be specified.
    rownames : sequence, default None
        If passed, must match number of row arrays passed.
    colnames : sequence, default None
        If passed, must match number of column arrays passed.
    aggfunc : function, optional
        If specified, requires `values` be specified as well.
    margins : bool, default False
        Add row/column margins (subtotals).
    margins_name : str, default 'All'
        Name of the row/column that will contain the totals
        when margins is True.
    dropna : bool, default True
        Do not include columns whose entries are all NaN.
    normalize : bool, {'all', 'index', 'columns'}, or {0,1}, default False
        Normalize by dividing all values by the sum of values.

        - If passed 'all' or `True`, will normalize over all values.
        - If passed 'index' will normalize over each row.
        - If passed 'columns' will normalize over each column.
        - If margins is `True`, will also normalize margin values.

    Returns
    -------
    DataFrame
        Cross tabulation of the data.

    See Also
    --------
    DataFrame.pivot : Reshape data based on column values.
    pivot_table : Create a pivot table as a DataFrame.

    Notes
    -----
    Any Series passed will have their name attributes used unless row or column
    names for the cross-tabulation are specified.

    Any input passed containing Categorical data will have **all** of its
    categories included in the cross-tabulation, even if the actual data does
    not contain any instances of a particular category.

    In the event that there aren't overlapping indexes an empty DataFrame will
    be returned.

    Reference :ref:`the user guide <reshaping.crosstabulations>` for more examples.

    Examples
    --------
    >>> a = np.array(["foo", "foo", "foo", "foo", "bar", "bar",
    ...               "bar", "bar", "foo", "foo", "foo"], dtype=object)
    >>> b = np.array(["one", "one", "one", "two", "one", "one",
    ...               "one", "two", "two", "two", "one"], dtype=object)
    >>> c = np.array(["dull", "dull", "shiny", "dull", "dull", "shiny",
    ...               "shiny", "dull", "shiny", "shiny", "shiny"],
    ...              dtype=object)
    >>> pd.crosstab(a, [b, c], rownames=['a'], colnames=['b', 'c'])
    b   one        two
    c   dull shiny dull shiny
    a
    bar    1     2    1     0
    foo    2     2    1     2

    Here 'c' and 'f' are not represented in the data and will not be
    shown in the output because dropna is True by default. Set
    dropna=False to preserve categories with no data.

    >>> foo = pd.Categorical(['a', 'b'], categories=['a', 'b', 'c'])
    >>> bar = pd.Categorical(['d', 'e'], categories=['d', 'e', 'f'])
    >>> pd.crosstab(foo, bar)
    col_0  d  e
    row_0
    a      1  0
    b      0  1
    >>> pd.crosstab(foo, bar, dropna=False)
    col_0  d  e  f
    row_0
    a      1  0  0
    b      0  1  0
    c      0  0  0
    """
    ...




"""
This type stub file was generated by pyright.
"""

import numpy as np
import pyarrow
from typing import Any, Callable, TYPE_CHECKING, TypeVar
from pandas._typing import npt
from pandas.util._decorators import cache_readonly
from pandas.core.arrays.masked import BaseMaskedArray, BaseMaskedDtype

if TYPE_CHECKING:
    ...
T = TypeVar("T", bound="NumericArray")
class NumericDtype(BaseMaskedDtype):
    _default_np_dtype: np.dtype
    _checker: Callable[[Any], bool]
    def __repr__(self) -> str:
        ...
    
    @cache_readonly
    def is_signed_integer(self) -> bool:
        ...
    
    @cache_readonly
    def is_unsigned_integer(self) -> bool:
        ...
    
    def __from_arrow__(self, array: pyarrow.Array | pyarrow.ChunkedArray) -> BaseMaskedArray:
        """
        Construct IntegerArray/FloatingArray from pyarrow Array/ChunkedArray.
        """
        ...
    


class NumericArray(BaseMaskedArray):
    """
    Base class for IntegerArray and FloatingArray.
    """
    _dtype_cls: type[NumericDtype]
    def __init__(self, values: np.ndarray, mask: npt.NDArray[np.bool_], copy: bool = ...) -> None:
        ...
    
    @cache_readonly
    def dtype(self) -> NumericDtype:
        ...
    
    _HANDLED_TYPES = ...





"""
This type stub file was generated by pyright.
"""

from typing import ChainMap, TypeVar

_KT = TypeVar("_KT")
_VT = TypeVar("_VT")
class DeepChainMap(ChainMap[_KT, _VT]):
    """
    Variant of ChainMap that allows direct updates to inner scopes.

    Only works when all passed mapping are mutable.
    """
    def __setitem__(self, key: _KT, value: _VT) -> None:
        ...
    
    def __delitem__(self, key: _KT) -> None:
        """
        Raises
        ------
        KeyError
            If `key` doesn't exist.
        """
        ...
    


